import copy
import gc
import glob
import inspect
import os
import tempfile
import warnings
from abc import ABC, ABCMeta, abstractmethod
from importlib.metadata import version
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import deepspeed
import dill
import numpy as np
import torch
from accelerate import Accelerator
from accelerate.utils import broadcast_object_list
from accelerate.utils.deepspeed import DeepSpeedOptimizerWrapper
from deepspeed.runtime.zero.config import DeepSpeedZeroConfig
from gymnasium import spaces
from numpy.typing import ArrayLike
from tensordict import TensorDict
from torch._dynamo import OptimizedModule

from agilerl.algorithms.core.registry import (
    HyperparameterConfig,
    MutationRegistry,
    NetworkGroup,
    OptimizerConfig,
)
from agilerl.algorithms.core.wrappers import OptimizerWrapper
from agilerl.protocols import (
    AgentWrapper,
    EvolvableAttributeDict,
    EvolvableAttributeType,
    EvolvableModule,
)
from agilerl.typing import (
    ActionType,
    ArrayDict,
    DeviceType,
    ExperiencesType,
    GymSpaceType,
    InfosDict,
    ObservationType,
    TorchObsType,
)
from agilerl.utils.algo_utils import (
    assert_supported_space,
    chkpt_attribute_to_device,
    clone_llm,
    compile_model,
    is_module_list,
    isroutine,
    key_in_nested_dict,
    preprocess_observation,
    recursive_check_module_attrs,
    remove_compile_prefix,
)
from agilerl.utils.evolvable_networks import is_image_space

__all__ = ["EvolvableAlgorithm", "RLAlgorithm", "MultiAgentRLAlgorithm"]

SelfEvolvableAlgorithm = TypeVar("SelfEvolvableAlgorithm", bound="EvolvableAlgorithm")
SelfRLAlgorithm = TypeVar("SelfRLAlgorithm", bound="RLAlgorithm")
SelfAgentWrapper = TypeVar("SelfAgentWrapper", bound="AgentWrapper")
MARLObservationType = Dict[str, ObservationType]


class _RegistryMeta(type):
    """Metaclass to wrap registry information after algorithm is done
    initializing with specified network groups and optimizers."""

    def __call__(
        cls: Type[SelfEvolvableAlgorithm], *args, **kwargs
    ) -> SelfEvolvableAlgorithm:
        # Create the instance
        instance: SelfEvolvableAlgorithm = super().__call__(*args, **kwargs)

        # Call the base class post_init_hook after all initialization
        if isinstance(instance, cls) and hasattr(instance, "_registry_init"):
            instance._registry_init()

        return instance


class RegistryMeta(_RegistryMeta, ABCMeta): ...


def get_checkpoint_dict(agent: SelfEvolvableAlgorithm) -> Dict[str, Any]:
    """Returns a dictionary of the agent's attributes to save in a checkpoint.

    :param agent: The agent to save.
    :type agent: EvolvableAlgorithm

    :return: A dictionary of the agent's attributes.
    :rtype: dict[str, Any]
    """
    attribute_dict = EvolvableAlgorithm.inspect_attributes(agent)

    # Extract info on evolvable modules and optimizers in the algorithm
    network_info: Dict[str, Dict[str, Any]] = {"modules": {}, "optimizers": {}}
    for attr in agent.evolvable_attributes():
        obj: EvolvableAttributeType = getattr(agent, attr)
        if isinstance(obj, OptimizerWrapper):
            network_info["optimizers"].update(
                {
                    f"{attr}_cls": obj.optimizer_cls.__name__,
                    f"{attr}_state_dict": obj.state_dict(),
                    f"{attr}_networks": obj.network_names,
                    f"{attr}_lr": obj.lr_name,
                    f"{attr}_kwargs": obj.optimizer_kwargs,
                    f"{attr}_multiagent": obj.multiagent,
                }
            )
        elif isinstance(obj, (OptimizedModule, EvolvableModule)) or is_module_list(obj):
            if is_module_list(obj):
                obj_list = obj
                obj_cls = [
                    (
                        m._orig_mod.__class__
                        if isinstance(m, OptimizedModule)
                        else m.__class__
                    )
                    for m in obj_list
                ]
                init_dict = [m.init_dict for m in obj_list]
                state_dict = [remove_compile_prefix(m.state_dict()) for m in obj_list]
            else:
                obj_list = [obj]
                obj_cls = (
                    obj._orig_mod.__class__
                    if isinstance(obj, OptimizedModule)
                    else obj.__class__
                )
                init_dict = obj.init_dict
                state_dict = remove_compile_prefix(obj.state_dict())

            network_info["modules"].update(
                {
                    f"{attr}_cls": obj_cls,
                    f"{attr}_init_dict": init_dict,
                    f"{attr}_state_dict": state_dict,
                }
            )
        else:
            raise TypeError(
                f"Something went wrong. Identified '{attr}' as an evolvable module "
                f"when it is of type {type(obj)}."
            )

    network_attr_names = [
        name for name in agent.evolvable_attributes(networks_only=True)
    ]
    optimizer_attr_names = [
        name
        for name in agent.evolvable_attributes()
        if isinstance(getattr(agent, name), OptimizerWrapper)
    ]

    network_info["network_names"] = network_attr_names
    network_info["optimizer_names"] = optimizer_attr_names
    attribute_dict["network_info"] = network_info
    attribute_dict["agilerl_version"] = version("agilerl")
    attribute_dict.pop("accelerator", None)
    if attribute_dict.pop("lr_scheduler", None) is not None:
        attribute_dict["lr_scheduler"] = agent.lr_scheduler.state_dict()
    return attribute_dict


class EvolvableAlgorithm(ABC, metaclass=RegistryMeta):
    """Base object for all algorithms in the AgileRL framework.

    :param index: The index of the individual.
    :type index: int
    :param hp_config: Hyperparameter configuration for the algorithm, defaults to None.
    :type hp_config: Optional[HyperparameterConfig], optional
    :param device: Device to run the algorithm on, defaults to "cpu".
    :type device: Union[str, torch.device], optional
    :param accelerator: Accelerator object for distributed computing, defaults to None.
    :type accelerator: Optional[Accelerator], optional
    :param torch_compiler: The torch compiler mode to use, defaults to None.
    :type torch_compiler: Optional[Any], optional
    :param name: Name of the algorithm, defaults to the class name.
    :type name: Optional[str], optional
    """

    def __init__(
        self,
        index: int,
        hp_config: Optional[HyperparameterConfig] = None,
        device: Union[str, torch.device] = "cpu",
        accelerator: Optional[Accelerator] = None,
        torch_compiler: Optional[Any] = None,
        name: Optional[str] = None,
    ) -> None:

        assert isinstance(index, int), "Agent index must be an integer."
        assert isinstance(device, (str, torch.device)), "Device must be a string."
        assert isinstance(name, (type(None), str)), "Name must be a string."
        assert isinstance(
            accelerator, (type(None), Accelerator)
        ), "Accelerator must be an instance of Accelerator."
        if torch_compiler:
            assert torch_compiler in [
                "default",
                "reduce-overhead",
                "max-autotune",
            ], "Choose between torch compiler modes: default, reduce-overhead, max-autotune or None"

        self.accelerator = accelerator
        self.device = device if self.accelerator is None else self.accelerator.device
        self.torch_compiler = torch_compiler
        self.algo = name if name is not None else self.__class__.__name__

        self._mut = None
        self._index = index
        self.scores = []
        self.fitness = []
        self.steps = [0]
        self.registry = MutationRegistry(hp_config)
        self.training = True

    @property
    def index(self) -> int:
        """Returns the index of the algorithm."""
        return self._index

    @index.setter
    def index(self, value: int) -> None:
        """Sets the index of the algorithm."""
        self._index = value

    @property
    def mut(self) -> Any:
        """Returns the mutation object of the algorithm."""
        return self._mut

    @mut.setter
    def mut(self, value: Optional[str]) -> None:
        """Sets the mutation object of the algorithm."""
        self._mut = value

    @abstractmethod
    def preprocess_observation(self, observation: ObservationType) -> TorchObsType:
        """Preprocesses observations for forward pass through neural network.

        :param observations: Observations of environment
        :type observations: numpy.ndarray[float] or dict[str, numpy.ndarray[float]]

        :return: Preprocessed observations
        :rtype: torch.Tensor[float] or dict[str, torch.Tensor[float]]
        """
        raise NotImplementedError

    @abstractmethod
    def learn(self, experiences: ExperiencesType, **kwargs) -> Any:
        """Abstract method for learning the algorithm."""
        raise NotImplementedError

    @abstractmethod
    def get_action(
        self, obs: Union[ObservationType, MARLObservationType], *args, **kwargs
    ) -> ActionType:
        """Abstract method for getting an action from the algorithm."""
        raise NotImplementedError

    @abstractmethod
    def test(self, *args, **kwargs) -> ArrayLike:
        """Abstract method for testing the algorithm."""
        raise NotImplementedError

    @staticmethod
    def get_state_dim(observation_space: GymSpaceType) -> Tuple[int, ...]:
        """Returns the dimension of the state space as it pertains to the underlying
        networks (i.e. the input size of the networks).

        :param observation_space: The observation space of the environment.
        :type observation_space: spaces.Space or List[spaces.Space].

        :return: The dimension of the state space.
        :rtype: Tuple[int, ...]."""
        if isinstance(observation_space, (list, tuple, spaces.Tuple)):
            assert_supported_space(observation_space)
            return tuple(
                EvolvableAlgorithm.get_state_dim(space) for space in observation_space
            )
        elif isinstance(observation_space, spaces.Dict):
            assert_supported_space(observation_space)
            return {
                key: EvolvableAlgorithm.get_state_dim(subspace)
                for key, subspace in observation_space.spaces.items()
            }
        elif isinstance(observation_space, spaces.Discrete):
            return (observation_space.n,)
        elif isinstance(observation_space, spaces.MultiDiscrete):
            return (sum(observation_space.nvec),)
        elif isinstance(observation_space, spaces.Box):
            return observation_space.shape
        elif isinstance(observation_space, spaces.MultiBinary):
            return (observation_space.n,)
        else:
            raise AttributeError(
                f"Can't access state dimensions for {type(observation_space)} spaces."
            )

    @staticmethod
    def get_action_dim(action_space: GymSpaceType) -> int:
        """Returns the dimension of the action space as it pertains to the underlying
        networks (i.e. the output size of the networks).

        :param action_space: The action space of the environment.
        :type action_space: spaces.Space or List[spaces.Space].

        :return: The dimension of the action space.
        :rtype: int.
        """
        if isinstance(action_space, (list, tuple)):
            return tuple(
                EvolvableAlgorithm.get_action_dim(space) for space in action_space
            )
        elif isinstance(action_space, spaces.MultiBinary):
            return action_space.n
        elif isinstance(action_space, spaces.Discrete):
            return action_space.n
        elif isinstance(action_space, spaces.MultiDiscrete):
            return sum(action_space.nvec)
        elif isinstance(action_space, spaces.Box):
            # NOTE: Here we assume the action space only has one dimension
            #       (i.e. the actions correspond to a one-dimensional vector)
            return action_space.shape[0]
        else:
            raise AttributeError(
                f"Can't access action dimensions for {type(action_space)} spaces."
            )

    @staticmethod
    def inspect_attributes(
        agent: SelfEvolvableAlgorithm, input_args_only: bool = False
    ) -> Dict[str, Any]:
        """
        Inspect and retrieve the attributes of the current object, excluding attributes related to the
        underlying evolvable networks (i.e. `EvolvableModule`, `torch.optim.Optimizer`) and with
        an option to include only the attributes that are input arguments to the constructor.

        :param input_args_only: If True, only include attributes that are input arguments to the constructor.
                                Defaults to False.
        :type input_args_only: bool
        :return: A dictionary of attribute names and their values.
        :rtype: dict[str, Any]
        """
        # Get all attributes of the current object
        attributes = inspect.getmembers(agent, lambda a: not isroutine(a))

        # Exclude attributes that are EvolvableModule or Optimizer objects (also check for nested
        # module-related attributes for multi-agent algorithms)
        exclude = list(agent.evolvable_attributes().keys())
        exclude += [attr for attr, val in attributes if isinstance(val, TensorDict)]

        # Exclude private and built-in attributes
        attributes = [
            a for a in attributes if not (a[0].startswith("_") or a[0].endswith("_"))
        ]

        # If input_args_only is True, only include attributes that are
        # input arguments to the constructor
        if input_args_only:
            constructor_params = inspect.signature(agent.__init__).parameters.keys()
            attributes = {
                k: v
                for k, v in attributes
                if k not in exclude and k in constructor_params
            }
        else:
            # Remove the algo specific guarded variables (if specified)
            attributes = {k: v for k, v in attributes if k not in exclude}
        return attributes

    @staticmethod
    def copy_attributes(
        agent: SelfEvolvableAlgorithm, clone: SelfEvolvableAlgorithm
    ) -> SelfEvolvableAlgorithm:
        """Copies the non-evolvable attributes of the algorithm to a clone.

        :param clone: The clone of the algorithm.
        :type clone: SelfEvolvableAlgorithm

        :return: The clone of the algorithm.
        :rtype: SelfEvolvableAlgorithm
        """
        for attribute in EvolvableAlgorithm.inspect_attributes(agent).keys():
            if hasattr(agent, attribute) and hasattr(clone, attribute):
                attr, clone_attr = getattr(agent, attribute), getattr(clone, attribute)

                # NOTE: Here we handle the case where the individual is wrapped by an
                # AgentWrapper object, which includes the agent itself and functools.partial
                # objects as attributes that shouldn't be copied
                if callable(attr) or isinstance(attr, EvolvableAlgorithm):
                    continue
                elif isinstance(attr, torch.Tensor) or isinstance(
                    clone_attr, torch.Tensor
                ):
                    if not torch.equal(attr, clone_attr):
                        try:
                            setattr(
                                clone,
                                attribute,
                                copy.deepcopy(getattr(agent, attribute)),
                            )
                        except RuntimeError:
                            # If the tensor is not a leaf tensor, we need to clone it using torch.clone
                            setattr(
                                clone, attribute, torch.clone(getattr(agent, attribute))
                            )

                elif isinstance(attr, np.ndarray) or isinstance(clone_attr, np.ndarray):
                    if not np.array_equal(attr, clone_attr):
                        setattr(
                            clone, attribute, copy.deepcopy(getattr(agent, attribute))
                        )
                elif isinstance(attr, list) or isinstance(clone_attr, list):
                    setattr(clone, attribute, [copy.deepcopy(el) for el in attr])
                elif isinstance(attr, dict) or isinstance(clone_attr, dict):
                    setattr(
                        clone,
                        attribute,
                        {key: copy.deepcopy(value) for key, value in attr.items()},
                    )
                elif attr != clone_attr or isinstance(attr, MutationRegistry):
                    setattr(clone, attribute, copy.deepcopy(getattr(agent, attribute)))
            else:
                setattr(clone, attribute, copy.deepcopy(getattr(agent, attribute)))
        return clone

    @classmethod
    def population(
        cls: Type[SelfEvolvableAlgorithm],
        size: int,
        observation_space: GymSpaceType,
        action_space: GymSpaceType,
        wrapper_cls: Optional[Type[SelfAgentWrapper]] = None,
        wrapper_kwargs: Dict[str, Any] = {},
        **kwargs,
    ) -> List[Union[SelfEvolvableAlgorithm, SelfAgentWrapper]]:
        """Creates a population of algorithms.

        :param size: The size of the population.
        :type size: int.

        :return: A list of algorithms.
        :rtype: List[SelfEvolvableAlgorithm].
        """
        if wrapper_cls is not None:
            return [
                wrapper_cls(
                    cls(observation_space, action_space, index=i, **kwargs),
                    **wrapper_kwargs,
                )
                for i in range(size)
            ]

        return [
            cls(observation_space, action_space, index=i, **kwargs) for i in range(size)
        ]

    def __setattr__(self, name: str, value: Any) -> None:
        """Sets the attribute of the algorithm. If the attribute is an OptimizerWrapper,
        we register the optimizer with the algorithms registry.

        :param name: The name of the attribute.
        :type name: str
        :param value: The value of the attribute.
        :type value: Any
        """
        if isinstance(value, OptimizerWrapper) and name not in [
            config.name for config in self.registry.optimizers
        ]:
            config = OptimizerConfig(
                name=name,
                networks=value.network_names,
                lr=value.lr_name,
                optimizer_cls=value.optimizer_cls,
                optimizer_kwargs=value.optimizer_kwargs,
                multiagent=value.multiagent,
            )
            self.registry.register_optimizer(config)

        super().__setattr__(name, value)

    def _registry_init(self) -> None:
        """Registers the networks, optimizers, and algorithm hyperparameters in the algorithm with
        the mutations registry. We also check that all of the evolvable networks and their respective
        optimizers have been registered with the algorithm, and that the user-specified hyperparameters
        to mutate have been set as attributes in the algorithm."""

        if not self.registry.groups:
            raise AttributeError(
                "No network groups have been registered in the algorithms __init__ method. "
                "Please register NetworkGroup objects specifying all of the evaluation and "
                "shared/target networks through the `register_network_group()` method."
            )

        # Check that all the inspected evolvable attributes can be found in the registry
        all_registered = self.registry.all_registered()
        not_found = [
            attr for attr in self.evolvable_attributes() if attr not in all_registered
        ]
        if not_found:
            raise AttributeError(
                f"The following evolvable attributes could not be found in the registry: {not_found}. "
                "Please check that the defined NetworkGroup objects contain all of the EvolvableModule objects "
                "in the algorithm."
            )

        # Check that one of the network groups relates to a policy
        if not any(group.policy for group in self.registry.groups):
            raise AttributeError(
                "No network group has been registered as a policy (i.e. the network used to "
                "select actions) in the registry. Please register a NetworkGroup object "
                "specifying the policy network."
            )

        # Check that all the hyperparameters to mutate have been set as attributes in the algorithm
        if self.registry.hp_config is not None:
            for hp in self.registry.hp_config:
                if not hasattr(self, hp):
                    raise AttributeError(
                        f"Hyperparameter {hp} was found in the mutations configuration but has "
                        "not been set as an attribute in the algorithm."
                    )

    def _wrap_attr(self, attr: EvolvableAttributeType) -> EvolvableModule:
        """Wraps the model with the accelerator.

        :param attr: The attribute to wrap.
        :type attr: EvolvableAttributeType

        :return: The wrapped attribute.
        :rtype: EvolvableModule
        """
        if isinstance(attr, OptimizerWrapper):
            if isinstance(attr.optimizer, list):
                wrapped_opt = [self.accelerator.prepare(opt) for opt in attr.optimizer]
            else:
                wrapped_opt = self.accelerator.prepare(attr.optimizer)

            attr.optimizer = wrapped_opt
            return attr

        # Only wrap the model if its part of the computation graph
        return self.accelerator.prepare(attr) if attr.state_dict() else attr

    def set_training_mode(self, training: bool) -> None:
        """Sets the training mode of the algorithm.

        :param training: If True, set the algorithm to training mode.
        :type training: bool
        """
        self.training = training

    def get_lr_names(self) -> List[str]:
        """Returns the learning rates of the algorithm."""
        return [opt.lr for opt in self.registry.optimizers]

    def register_network_group(self, group: NetworkGroup) -> None:
        """Sets the evaluation network for the algorithm.

        :param name: The name of the evaluation network.
        :type name: str
        """
        self.registry.register_group(group)

    def register_mutation_hook(self, hook: Callable) -> None:
        """Registers a hook to be executed after a mutation is performed on
        the algorithm.

        :param hook: The hook to be executed after mutation.
        :type hook: Callable
        """
        self.registry.register_hook(hook)

    def mutation_hook(self) -> None:
        """Executes the hooks registered with the algorithm."""
        for hook in self.registry.hooks:
            getattr(self, hook)()

    def get_policy(self) -> EvolvableModule:
        """Returns the policy network of the algorithm."""
        for group in self.registry.groups:
            if group.policy:
                return getattr(self, group.eval)

        raise AttributeError(
            "No policy network has been registered with the algorithm."
        )

    def recompile(self) -> None:
        """Recompiles the evolvable modules in the algorithm with the specified torch compiler."""
        for name, obj in self.evolvable_attributes(networks_only=True).items():
            if isinstance(obj, list):
                compiled = [
                    compile_model(module, self.torch_compiler) for module in obj
                ]
            else:
                compiled = compile_model(obj, self.torch_compiler)

            setattr(self, name, compiled)

    def to_device(self, *experiences: TorchObsType) -> Tuple[TorchObsType, ...]:
        """Moves experiences to the device.

        :param experiences: Experiences to move to device
        :type experiences: Tuple[torch.Tensor[float], ...]

        :return: Experiences on the device
        :rtype: Tuple[torch.Tensor[float], ...]
        """
        device = self.device if self.accelerator is None else self.accelerator.device
        on_device = []
        for exp in experiences:
            if isinstance(exp, dict):
                exp = {key: val.to(device) for key, val in exp.items()}
            elif isinstance(exp, (list, tuple)) and isinstance(exp[0], torch.Tensor):
                exp = tuple(val.to(device) for val in exp)
            elif isinstance(exp, torch.Tensor):
                exp = exp.to(device)

            on_device.append(exp)

        return on_device

    def evolvable_attributes(
        self, networks_only: bool = False
    ) -> EvolvableAttributeDict:
        """Returns the attributes related to the evolvable networks in the algorithm. Includes
        attributes that are either evolvable networks or a list of evolvable networks, as well
        as the optimizers associated with the networks.

        :param networks_only: If True, only include evolvable networks, defaults to False
        :type networks_only: bool, optional

        :return: A dictionary of network attributes.
        :rtype: dict[str, Any]
        """

        def is_evolvable(attr: str, obj: Any):
            return (
                recursive_check_module_attrs(obj, networks_only)
                and not attr.startswith("_")
                and not attr.endswith("_")
            )

        # Inspect evolvable given specs
        evolvable_attrs = {}
        for attr in dir(self):
            obj = getattr(self, attr)
            if is_evolvable(attr, obj):
                evolvable_attrs[attr] = obj

        return evolvable_attrs

    def wrap_models(self) -> None:
        """Wraps the models in the algorithm with the accelerator."""
        if self.accelerator is None:
            return

        for attr in self.evolvable_attributes():
            obj = getattr(self, attr)
            if isinstance(obj, list):
                setattr(self, attr, [self._wrap_attr(m) for m in obj])
            else:
                setattr(self, attr, self._wrap_attr(obj))

    def unwrap_models(self):
        """Unwraps the models in the algorithm from the accelerator."""
        if self.accelerator is None:
            raise AttributeError("No accelerator has been set for the algorithm.")

        for attr in self.evolvable_attributes(networks_only=True):
            obj = getattr(self, attr)
            if isinstance(obj, list):
                setattr(self, attr, [self.accelerator.unwrap_model(m) for m in obj])
            else:
                setattr(self, attr, self.accelerator.unwrap_model(obj))

    def clone(
        self: SelfEvolvableAlgorithm, index: Optional[int] = None, wrap: bool = True
    ) -> SelfEvolvableAlgorithm:
        """Creates a clone of the algorithm.

        :param index: The index of the clone, defaults to None
        :type index: Optional[int], optional
        :param wrap: If True, wrap the models in the clone with the accelerator, defaults to False
        :type wrap: bool, optional

        :return: A clone of the algorithm
        :rtype: EvolvableAlgorithm
        """
        # Make copy using input arguments
        input_args = EvolvableAlgorithm.inspect_attributes(self, input_args_only=True)
        input_args["wrap"] = wrap

        clone = type(self)(**input_args)

        if self.accelerator is not None:
            self.unwrap_models()

        # Clone evolvable modules
        cloned_modules = {}
        for attr, obj in self.evolvable_attributes(networks_only=True).items():
            if isinstance(obj, list):
                cloned_modules[attr] = [m.clone() for m in obj]
            else:
                cloned_modules[attr] = obj.clone()

            setattr(clone, attr, cloned_modules[attr])

        # Run mutation hook at this step given possibility of sharing
        # encoder parameters between networks
        clone.mutation_hook()

        # Reinitialize optimizers
        for opt_config in self.registry.optimizers:
            orig_optimizer: OptimizerWrapper = getattr(self, opt_config.name)

            networks = (
                cloned_modules[opt_config.networks[0]]
                if opt_config.multiagent
                else [cloned_modules[net] for net in opt_config.networks]
            )
            opt = OptimizerWrapper(
                getattr(torch.optim, opt_config.optimizer_cls),
                networks=networks,
                lr=orig_optimizer.lr,
                network_names=opt_config.networks,
                lr_name=opt_config.lr,
                optimizer_kwargs=opt_config.optimizer_kwargs,
                multiagent=opt_config.multiagent,
            )
            opt.load_state_dict(orig_optimizer.state_dict())
            setattr(clone, opt_config.name, opt)

        # Prepare with accelerator / compiler if necessary
        if self.accelerator is not None and wrap:
            clone.wrap_models()
        elif self.torch_compiler:
            torch.set_float32_matmul_precision("high")
            clone.recompile()

        # Copy non-evolvable attributes back to clone
        clone = EvolvableAlgorithm.copy_attributes(self, clone)
        if index is not None:
            clone.index = index

        return clone

    def save_checkpoint(self, path: str) -> None:
        """Saves a checkpoint of agent properties and network weights to path.

        :param path: Location to save checkpoint at
        :type path: string
        """
        torch.save(
            get_checkpoint_dict(self),
            path,
            pickle_module=dill,
        )

    def load_checkpoint(self, path: str) -> None:
        """Loads saved agent properties and network weights from checkpoint.

        :param path: Location to load checkpoint from
        :type path: string
        """
        checkpoint: Dict[str, Any] = torch.load(
            path, map_location=self.device, pickle_module=dill
        )

        # Recreate evolvable modules
        network_info: Dict[str, Dict[str, Any]] = checkpoint["network_info"]
        network_names = network_info["network_names"]
        for name in network_names:
            net_dict = {
                k: v for k, v in network_info["modules"].items() if k.startswith(name)
            }

            module_cls = net_dict[f"{name}_cls"]
            init_dict = net_dict[f"{name}_init_dict"]
            if isinstance(module_cls, list):
                loaded_modules = []
                for mod, d in zip(module_cls, init_dict):
                    loaded_mod: EvolvableModule = mod(**d)
                    loaded_modules.append(loaded_mod)

                setattr(self, name, loaded_modules)
            else:
                loaded_module: EvolvableModule = module_cls(**init_dict)
                setattr(self, name, loaded_module)

        # Apply mutation hooks
        # NOTE: We do this before loading the state dicts because there may be
        # hooks that pertain to the network parameters such as e.g. encoder parameter
        # sharing
        self.mutation_hook()

        # Load state dicts after applying mutation hook
        for name in network_names:
            net_dict = {
                k: v for k, v in network_info["modules"].items() if k.startswith(name)
            }
            loaded_module = getattr(self, name)
            state_dict = net_dict[f"{name}_state_dict"]
            if isinstance(loaded_module, list):
                for loaded_mod, state in zip(loaded_module, state_dict):
                    if state:
                        loaded_mod.load_state_dict(state)
            elif state_dict:
                loaded_module.load_state_dict(state_dict)

        optimizer_names = network_info["optimizer_names"]
        for name in optimizer_names:
            opt_dict = {
                k: v
                for k, v in network_info["optimizers"].items()
                if k.startswith(name)
            }

            # Initialize optimizer
            opt_kwargs = opt_dict[f"{name}_kwargs"]
            optimizer_cls = opt_dict[f"{name}_cls"]
            opt_networks = opt_dict[f"{name}_networks"]
            opt_lr = opt_dict[f"{name}_lr"]
            is_multiagent = opt_dict[f"{name}_multiagent"]
            networks = (
                getattr(self, opt_networks[0])
                if is_multiagent
                else [getattr(self, net) for net in opt_networks]
            )
            optimizer = OptimizerWrapper(
                getattr(torch.optim, optimizer_cls),
                networks=networks,
                lr=getattr(self, opt_lr),
                network_names=opt_networks,
                lr_name=opt_lr,
                optimizer_kwargs=opt_kwargs,
                multiagent=is_multiagent,
            )

            # Load optimizer state
            optimizer.load_state_dict(opt_dict[f"{name}_state_dict"])
            setattr(self, name, optimizer)

        # Check loaded registry is consistent with the algorithm
        if checkpoint["registry"] != self.registry:
            raise ValueError(
                "Loaded registry does not match the algorithm's registry. Please make "
                "sure you are loading the checkpoint with the correct algorithm."
            )

        # Load other attributes
        checkpoint.pop("network_info")
        for attribute in checkpoint.keys():
            setattr(self, attribute, checkpoint[attribute])

        # Wrap models / compile if necessary
        if self.accelerator is not None:
            self.wrap_models()
        elif self.torch_compiler:
            torch.set_float32_matmul_precision("high")
            self.recompile()

    @classmethod
    def load(
        cls: Type[SelfEvolvableAlgorithm],
        path: str,
        device: DeviceType = "cpu",
        accelerator: Optional[Accelerator] = None,
    ) -> SelfEvolvableAlgorithm:
        """Loads an algorithm from a checkpoint.

        :param path: Location to load checkpoint from.
        :type path: string
        :param device: Device to load the algorithm on, defaults to 'cpu'
        :type device: str, optional
        :param accelerator: Accelerator object for distributed computing, defaults to None
        :type accelerator: Optional[Accelerator], optional

        :return: An instance of the algorithm
        :rtype: RLAlgorithm
        """
        checkpoint: Dict[str, Any] = torch.load(
            path, map_location=device, pickle_module=dill
        )

        # Reconstruct evolvable modules in algorithm
        network_info: Optional[Dict[str, Dict[str, Any]]] = checkpoint.get(
            "network_info"
        )
        if network_info is None:
            raise ValueError(
                "Network info not found in checkpoint. You may be loading a checkpoint from "
                "an older version of AgileRL. Since v2.0, we require AgileRL algorithms to "
                "have a specific structure to simplify evolutionary hyperparameter optimization. "
                "Please downgrade to v1.0.30 to load checkpoints from before this change."
            )

        network_names = network_info["network_names"]
        loaded_modules: Dict[str, EvolvableAttributeType] = {}
        for name in network_names:
            net_dict = {
                k: v for k, v in network_info["modules"].items() if k.startswith(name)
            }

            # Add device to init dict
            init_dict = net_dict.get(f"{name}_init_dict", None)
            if init_dict is None:
                raise ValueError(f"Init dict for {name} not found in checkpoint.")

            init_dict = chkpt_attribute_to_device(init_dict, device)

            # Reconstruct the modules
            module_cls: Union[Type[EvolvableModule], List[Type[EvolvableModule]]] = (
                net_dict[f"{name}_cls"]
            )
            if isinstance(module_cls, list):
                loaded_modules[name] = []
                for mod_cls, d in zip(module_cls, init_dict):
                    d["device"] = device
                    mod: EvolvableModule = mod_cls(**d)
                    loaded_modules[name].append(mod)
            else:
                init_dict["device"] = device
                module = module_cls(**init_dict)
                loaded_modules[name] = module

        # Reconstruct the algorithm
        constructor_params = inspect.signature(cls.__init__).parameters.keys()
        class_init_dict = {
            k: v for k, v in checkpoint.items() if k in constructor_params
        }

        checkpoint["accelerator"] = accelerator
        checkpoint["device"] = device
        self = cls(**class_init_dict)
        registry: MutationRegistry = checkpoint["registry"]
        self.registry = registry

        # Set loaded modules
        for name, module in loaded_modules.items():
            setattr(self, name, module)

        # Apply mutation hooks
        self.mutation_hook()

        # Load state dictionaries
        for name in network_names:
            net_dict = {
                k: v for k, v in network_info["modules"].items() if k.startswith(name)
            }
            loaded_module: Union[EvolvableModule, List[EvolvableModule]] = getattr(
                self, name
            )
            state_dict = net_dict[f"{name}_state_dict"]
            if isinstance(loaded_module, list):
                for loaded_mod, state in zip(loaded_module, state_dict):
                    if state:
                        loaded_mod.load_state_dict(state)
            elif state_dict:
                loaded_module.load_state_dict(state_dict)

        # Reconstruct optimizers in algorithm
        optimizer_names = network_info["optimizer_names"]
        loaded_optimizers = {}
        for name in optimizer_names:
            opt_dict = {
                k: v
                for k, v in network_info["optimizers"].items()
                if k.startswith(name)
            }

            # Add device to optimizer kwargs
            opt_kwargs = chkpt_attribute_to_device(opt_dict[f"{name}_kwargs"], device)
            lr = opt_dict[f"{name}_lr"]
            optimizer_cls = opt_dict[f"{name}_cls"]
            opt_networks = opt_dict[f"{name}_networks"]

            networks = (
                loaded_modules[opt_networks[0]]
                if opt_dict[f"{name}_multiagent"]
                else [loaded_modules[net] for net in opt_networks]
            )

            optimizer = OptimizerWrapper(
                getattr(torch.optim, optimizer_cls),
                networks=networks,
                lr=getattr(self, lr),
                network_names=opt_networks,
                lr_name=lr,
                optimizer_kwargs=opt_kwargs,
                multiagent=opt_dict[f"{name}_multiagent"],
            )

            state_dict = chkpt_attribute_to_device(
                opt_dict[f"{name}_state_dict"], device
            )
            optimizer.load_state_dict(state_dict)
            loaded_optimizers[name] = optimizer

        # Assign loaded modules and optimizers to the algorithm
        for name, module in loaded_modules.items():
            setattr(self, name, module)

        for name, optimizer in loaded_optimizers.items():
            setattr(self, name, optimizer)

        # Assign other attributes to the algorithm
        for attribute in EvolvableAlgorithm.inspect_attributes(self).keys():
            if attribute not in checkpoint:
                warnings.warn(
                    f"Attribute {attribute} not found in checkpoint. Skipping."
                )
                continue

            setattr(self, attribute, checkpoint.get(attribute))

        # Wrap models / compile if necessary
        if accelerator is not None:
            self.wrap_models()
        elif self.torch_compiler:
            torch.set_float32_matmul_precision("high")
            self.recompile()

        # Check for agent wrapper
        wrapper_cls = checkpoint.get("wrapper_cls")
        if wrapper_cls is not None:
            init_dict = checkpoint.get("wrapper_init_dict")
            wrapper_attributes = checkpoint.get("wrapper_attrs")
            self = wrapper_cls(self, **init_dict)
            for attr in wrapper_attributes:
                setattr(self, attr, wrapper_attributes[attr])

        return self


class RLAlgorithm(EvolvableAlgorithm, ABC):
    """Base object for all single-agent algorithms in the AgileRL framework.

    :param observation_space: The observation space of the environment.
    :type observation_space: spaces.Space
    :param action_space: The action space of the environment.
    :type action_space: spaces.Space
    :param index: The index of the individual.
    :type index: int
    :param learn_step: Learning frequency, defaults to 2048.
    :type learn_step: int, optional
    :param device: Device to run the algorithm on, defaults to "cpu".
    :type device: Union[str, torch.device], optional
    :param accelerator: Accelerator object for distributed computing, defaults to None.
    :type accelerator: Optional[Accelerator], optional
    :param normalize_images: If True, normalize images, defaults to True.
    :type normalize_images: bool, optional
    :param name: Name of the algorithm, defaults to the class name.
    :type name: Optional[str], optional
    """

    def __init__(
        self,
        observation_space: spaces.Space,
        action_space: spaces.Space,
        index: int,
        hp_config: Optional[HyperparameterConfig] = None,
        device: Union[str, torch.device] = "cpu",
        accelerator: Optional[Accelerator] = None,
        torch_compiler: Optional[Any] = None,
        normalize_images: bool = True,
        name: Optional[str] = None,
    ) -> None:

        super().__init__(index, hp_config, device, accelerator, torch_compiler, name)

        assert isinstance(
            observation_space, spaces.Space
        ), "Observation space must be an instance of gymnasium.spaces.Space."
        assert isinstance(
            action_space, spaces.Space
        ), "Action space must be an instance of gymnasium.spaces.Space."

        self.observation_space = observation_space
        self.action_space = action_space
        self.normalize_images = normalize_images

        # TODO: Temporary hack to support legacy code
        self.state_dim = self.get_state_dim(observation_space)
        self.action_dim = self.get_action_dim(action_space)
        self.discrete_actions = isinstance(
            action_space, (spaces.Discrete, spaces.MultiDiscrete)
        )
        self.min_action = (
            np.array(action_space.low).astype(np.float32)
            if hasattr(action_space, "low")
            else None
        )
        self.max_action = (
            np.array(action_space.high).astype(np.float32)
            if hasattr(action_space, "high")
            else None
        )

    def preprocess_observation(self, observation: ObservationType) -> TorchObsType:
        """Preprocesses observations for forward pass through neural network.

        :param observations: Observations of environment
        :type observations: ObservationType

        :return: Preprocessed observations
        :rtype: torch.Tensor[float] or dict[str, torch.Tensor[float]] or Tuple[torch.Tensor[float], ...]
        """
        return preprocess_observation(
            self.observation_space,
            observation=observation,
            device=self.device,
            normalize_images=self.normalize_images,
        )


class MultiAgentRLAlgorithm(EvolvableAlgorithm, ABC):
    """Base object for all multi-agent algorithms in the AgileRL framework.

    :param observation_spaces: The observation spaces of the agent environments.
    :type observation_spaces: List[spaces.Space]
    :param action_spaces: The action spaces of the agent environments.
    :type action_spaces: List[spaces.Space]
    :param agent_ids: The agent IDs of the agents in the environment.
    :type agent_ids: List[int]
    :param index: The index of the individual in the population.
    :type index: int.
    :param learn_step: Learning frequency, defaults to 2048
    :type learn_step: int, optional
    :param device: Device to run the algorithm on, defaults to "cpu"
    :type device: str, optional
    :param accelerator: Accelerator object for distributed computing, defaults to None
    :type accelerator: Optional[Accelerator], optional
    :param torch_compiler: The torch compiler mode to use, defaults to None
    :type torch_compiler: Optional[Any], optional
    :param normalize_images: If True, normalize images, defaults to True
    :type normalize_images: bool, optional
    :param placeholder_value: The value to use as placeholder for missing observations, defaults to -1.
    :type placeholder_value: Optional[Any], optional
    :param name: Name of the algorithm, defaults to the class name
    :type name: Optional[str], optional
    """

    def __init__(
        self,
        observation_spaces: Iterable[spaces.Space],
        action_spaces: Iterable[spaces.Space],
        agent_ids: Iterable[int],
        index: int,
        hp_config: Optional[HyperparameterConfig] = None,
        device: Union[str, torch.device] = "cpu",
        accelerator: Optional[Accelerator] = None,
        torch_compiler: Optional[Any] = None,
        normalize_images: bool = True,
        placeholder_value: Optional[Any] = -1,
        name: Optional[str] = None,
    ) -> None:

        super().__init__(index, hp_config, device, accelerator, torch_compiler, name)

        assert isinstance(
            agent_ids, (tuple, list)
        ), "Agent IDs must be stores in a tuple or list."
        assert len(agent_ids) == len(
            observation_spaces
        ), "Number of agent IDs must match number of observation spaces."
        assert isinstance(
            observation_spaces, (list, tuple)
        ), "Observation spaces must be a list or tuple."
        assert all(
            isinstance(_space, spaces.Space) for _space in observation_spaces
        ), "Observation spaces must be instances of gymnasium.spaces.Space."
        assert isinstance(
            action_spaces, (list, tuple)
        ), "Action spaces must be a list or tuple."
        assert all(
            isinstance(_space, spaces.Space) for _space in action_spaces
        ), "Action spaces must be instances of gymnasium.spaces.Space."

        if not all(
            isinstance(space, observation_spaces[0].__class__)
            for space in observation_spaces
        ):
            raise ValueError(
                "AgileRL 2.0 only supports homogeneous multi-agent environments "
                "(i.e. all agents have the same type of observation space)"
            )

        # TODO: This can be removed once we have a more robust implementation
        self.state_dims = self.get_state_dim(observation_spaces)
        self.action_dims = self.get_action_dim(action_spaces)
        self.one_hot = all(
            isinstance(space, (spaces.Discrete, spaces.MultiDiscrete))
            for space in observation_spaces
        )
        self.discrete_actions = all(
            isinstance(space, (spaces.Discrete, spaces.MultiDiscrete))
            for space in action_spaces
        )

        # For continuous action spaces, store the min and max action values
        if all(isinstance(space, spaces.Box) for space in action_spaces):
            self.min_action = [
                np.array(space.low).astype(np.float32) for space in action_spaces
            ]
            self.max_action = [
                np.array(space.high).astype(np.float32) for space in action_spaces
            ]
        else:
            self.min_action, self.max_action = None, None

        self.agent_ids = agent_ids
        self.n_agents = len(agent_ids)
        self.placeholder_value = placeholder_value

        self.shared_agent_ids = []
        self.homogeneous_agents = {}
        self.unique_observation_spaces = {}
        self.unique_action_spaces = {}
        for agent_id, obs_space, action_space in zip(
            self.agent_ids, observation_spaces, action_spaces
        ):
            # Split agent names on expected pattern of e.g. speaker_0, speaker_1,
            # listener_0, listener_1, to determine which agents are homogeneous
            homo_id = self.get_homo_id(agent_id)
            if homo_id in self.homogeneous_agents:
                self.homogeneous_agents[homo_id].append(agent_id)
                assert obs_space == self.unique_observation_spaces[homo_id], (
                    f"Homogeneous agents, i.e. agents that share the prefix {homo_id}, "
                    f"must have the same observation space. Found {self.unique_observation_spaces[homo_id]} and {obs_space}."
                )
                assert action_space == self.unique_action_spaces[homo_id], (
                    f"Homogeneous agents, i.e. agents that share the prefix {homo_id}, "
                    f"must have the same action space. Found {self.unique_action_spaces[homo_id]} and {action_space}."
                )
            else:
                self.shared_agent_ids.append(homo_id)
                self.homogeneous_agents[homo_id] = [agent_id]
                self.unique_observation_spaces[homo_id] = obs_space
                self.unique_action_spaces[homo_id] = action_space

        self.n_unique_agents = len(self.shared_agent_ids)
        self.normalize_images = normalize_images
        self.observation_spaces = observation_spaces
        self.action_spaces = action_spaces
        self.total_actions = sum(self.action_dims)
        self.total_state_dims = None

        if not any(
            isinstance(space, (spaces.Dict, spaces.Tuple))
            for space in observation_spaces
        ):
            self.total_state_dims = sum(state_dim[0] for state_dim in self.state_dims)

        # Build observation and action space dictionaries using agent IDs
        self.single_space = observation_spaces[0]
        self.observation_space = spaces.Dict(
            {agent_id: space for agent_id, space in zip(agent_ids, observation_spaces)}
        )
        self.action_space = spaces.Dict(
            {agent_id: space for agent_id, space in zip(agent_ids, action_spaces)}
        )

    def get_homo_id(self, agent_id: str) -> str:
        """Get the homogeneous ID for an agent.

        :param agent_id: The agent ID
        :type agent_id: str
        :return: The homogeneous ID
        """
        return agent_id.rsplit("_", 1)[0] if isinstance(agent_id, str) else agent_id

    def preprocess_observation(
        self, observation: ObservationType
    ) -> Dict[str, TorchObsType]:
        """Preprocesses observations for forward pass through neural network.

        :param observations: Observations of environment
        :type observations: numpy.ndarray[float] or dict[str, numpy.ndarray[float]]

        :return: Preprocessed observations
        :rtype: torch.Tensor[float] or dict[str, torch.Tensor[float]] or Tuple[torch.Tensor[float], ...]
        """
        preprocessed = {}
        for agent_id, agent_obs in observation.items():
            preprocessed[agent_id] = preprocess_observation(
                self.observation_space.get(agent_id),
                observation=agent_obs,
                device=self.device,
                normalize_images=self.normalize_images,
                placeholder_value=self.placeholder_value,
            )

        return preprocessed

    def extract_action_masks(self, infos: InfosDict) -> ArrayDict:
        """Extract action masks from info dictionary

        :param infos: Info dict
        :type infos: Dict[str, Dict[...]]

        :return: Action masks
        :rtype: Dict[str, np.ndarray]
        """
        action_masks = {
            agent: info.get("action_mask", None) if isinstance(info, dict) else None
            for agent, info in infos.items()
            if agent in self.agent_ids
        }  # Get dict of form {"agent_id" : [1, 0, 0, 0]...} etc

        return action_masks

    def extract_agent_masks(self, infos: InfosDict) -> Tuple[ArrayDict, ArrayDict]:
        """Extract env_defined_actions from info dictionary and determine agent masks

        :param infos: Info dict
        :type infos: Dict[str, Dict[...]]
        """

        # Deal with case of no env_defined_actions defined in the info dict
        # Deal with empty info dicts for each sub agent
        if not key_in_nested_dict(infos, "env_defined_actions") or all(
            not info for agent, info in infos.items() if agent in self.agent_ids
        ):
            return None, None

        env_defined_actions = {
            agent: (
                info.get("env_defined_actions", None)
                if isinstance(info, dict)
                else None
            )
            for agent, info in infos.items()
            if agent in self.agent_ids
        }
        agent_masks = None
        if env_defined_actions is not None:
            agent_masks = {}
            for idx, agent in enumerate(env_defined_actions.keys()):
                # Handle None if environment isn't vectorized
                if env_defined_actions[agent] is None:
                    if not self.discrete_actions:
                        nan_arr = np.empty(self.action_dims[idx])
                        nan_arr[:] = np.nan
                    else:
                        nan_arr = np.array([[np.nan]])
                    env_defined_actions[agent] = nan_arr

                # Handle discrete actions + env not vectorized
                if isinstance(env_defined_actions[agent], (int, float)):
                    env_defined_actions[agent] = np.array(
                        [[env_defined_actions[agent]]]
                    )

                # Ensure additional dimension is added in so shapes align for masking
                if len(env_defined_actions[agent].shape) == 1:
                    env_defined_actions[agent] = (
                        env_defined_actions[agent][:, np.newaxis]
                        if self.discrete_actions
                        else env_defined_actions[agent][np.newaxis, :]
                    )
                agent_masks[agent] = np.where(
                    np.isnan(env_defined_actions[agent]), 0, 1
                ).astype(bool)

        return env_defined_actions, agent_masks

    def stack_critic_observations(self, obs: Dict[str, TorchObsType]) -> TorchObsType:
        """Process observations for critic network input.

        .. note:: Assumes that the observation spaces for the different agents are the same.

        :param obs: Observation dict
        :type obs: Dict[str, torch.Tensor]

        :return: Stacked observations
        :rtype: torch.Tensor
        """
        obs = list(obs.values())
        if isinstance(self.single_space, spaces.Dict):
            processed_obs = {}
            for key, space in self.single_space.spaces.items():
                if is_image_space(space):
                    processed_obs[key] = torch.stack(
                        [obs[i][key] for i in range(self.n_agents)], dim=2
                    )
                else:
                    processed_obs[key] = torch.cat(
                        [obs[i][key] for i in range(self.n_agents)], dim=1
                    )

        elif isinstance(self.single_space, spaces.Tuple):
            processed_obs = []
            for i, space in enumerate(self.single_space):
                if is_image_space(space):
                    processed_obs.append(
                        torch.stack([obs[j][i] for j in range(self.n_agents)], dim=2)
                    )
                else:
                    processed_obs.append(
                        torch.cat([obs[j][i] for j in range(self.n_agents)], dim=1)
                    )
            processed_obs = tuple(processed_obs)

        elif is_image_space(self.single_space):
            processed_obs = torch.stack(obs, dim=2)
        else:
            processed_obs = torch.cat(obs, dim=1)

        return processed_obs

    def extract_inactive_agents(
        self, obs: Dict[str, ObservationType]
    ) -> Tuple[Dict[str, np.ndarray], Dict[str, ObservationType]]:
        """Extract inactive agents from observation.

        :param obs: Observation dictionary
        :type obs: Dict[str, ObservationType]

        :return: Tuple of inactive agents and filtered observations
        :rtype: Tuple[Dict[str, np.ndarray], Dict[str, ObservationType]]
        """
        inactive_agents = {}
        agents_to_remove = []
        for agent_id, agent_obs in obs.items():
            if isinstance(agent_obs, dict):
                sample = next(iter(agent_obs.values()))
            elif isinstance(agent_obs, tuple):
                sample = agent_obs[0]
            else:
                sample = agent_obs

            # For non-vectorised environments we assume observations for
            # inactive agents are not returned
            if len(sample.shape) == 1:
                continue

            # Extract rows where all values are -1
            inactive_agent_indices = np.where(np.isnan(sample).all(axis=1))[0]

            # If agent is active in all environments, don't need to save info
            if inactive_agent_indices.shape[0] == 0:
                continue

            filtered_obs: np.ndarray = np.delete(
                obs[agent_id], inactive_agent_indices, axis=0
            )

            # Don't need to save info if same agent is inactive in all environments
            if filtered_obs.shape[0] == 0:
                agents_to_remove.append(agent_id)
                continue

            obs[agent_id] = filtered_obs
            inactive_agents[agent_id] = inactive_agent_indices

        for agent_id in agents_to_remove:
            obs.pop(agent_id)

        return inactive_agents, obs

    def disassemble_homogeneous_outputs(
        self,
        homo_outputs: ArrayDict,
        vect_dim: int,
        homogeneous_agents: Dict[str, List[str]],
    ) -> ArrayDict:
        """Disassembles batched output by shared policies into their homogeneous agents' outputs.

        .. note:: This assumes that for any given sub-agent the termination condition is deterministic,
            i.e. any given agent will always terminate at the same timestep in different vectorized environments.

        :param homo_outputs: Dictionary to be disassembled, has the form {'agent': [4, 7, 8]}
        :type homo_outputs: Dict[str, np.ndarray]
        :param vect_dim: Vectorization dimension size, i.e. number of vect envs
        :type vect_dim: int
        :param homogeneous_agents: Dictionary of homogeneous agent IDs
        :type homogeneous_agents: Dict[str, List[str]]
        :return: Assembled dictionary, e.g. {'agent_0': 4, 'agent_1': 7, 'agent_2': 8}
        :rtype: Dict[str, np.ndarray]
        """
        output_dict = {}
        for homo_id, agent_ids in homogeneous_agents.items():
            homo_outputs[homo_id] = np.reshape(
                homo_outputs[homo_id],
                (len(agent_ids), vect_dim, -1),
            )
            for i, agent_id in enumerate(agent_ids):
                output_dict[agent_id] = homo_outputs[homo_id][i]

        return output_dict

    def sum_shared_rewards(self, rewards: ArrayDict) -> ArrayDict:
        """Sums the rewards for homogeneous agents

        :param rewards: Reward dictionary from environment
        :type rewards: Dict[str, np.ndarray]
        :return: Summed rewards dictionary
        :rtype: Dict[str, np.ndarray]
        """
        reward_shape = list(rewards.values())[0]
        reward_shape = (
            reward_shape.shape if isinstance(reward_shape, np.ndarray) else (1,)
        )
        summed_rewards = {
            agent_id: np.zeros(reward_shape) for agent_id in self.shared_agent_ids
        }
        for agent_id, reward in rewards.items():
            homo_id = self.get_homo_id(agent_id)
            summed_rewards[homo_id] += reward

        return summed_rewards

    def assemble_homogeneous_outputs(
        self, agent_outputs: ArrayDict, vect_dim: int
    ) -> ArrayDict:
        """Assembles individual agent outputs into batched outputs for shared policies.

        :param agent_outputs: Dictionary with individual agent outputs, e.g. {'agent_0': 4, 'agent_1': 7, 'agent_2': 8}
        :type agent_outputs: Dict[str, np.ndarray]
        :param vect_dim: Vectorization dimension size, i.e. number of vect envs
        :type vect_dim: int
        :return: Assembled dictionary with the form {'agent': [4, 7, 8]}
        :rtype: Dict[str, np.ndarray]
        """
        homo_outputs = {}
        for unique_id in self.shared_agent_ids:
            # Get all outputs for agents that share this ID
            homo_agent_outputs = []
            for homo_id in self.homogeneous_agents[unique_id]:
                if homo_id in agent_outputs:
                    homo_agent_outputs.append(agent_outputs[homo_id])

            if homo_agent_outputs:
                # Stack outputs along first dimension
                stacked_outputs = np.stack(homo_agent_outputs, axis=0)
                # Reshape into a form suitable for batch processing
                homo_outputs[unique_id] = np.reshape(
                    stacked_outputs, (len(homo_agent_outputs) * vect_dim, -1)
                )

        return homo_outputs


class LLMAlgorithm(EvolvableAlgorithm, ABC):
    """Base object for all LLM algorithms in the AgileRL framework."""

    def __init__(
        self,
        observation_space: spaces.Space,
        action_space: spaces.Space,
        index: int,
        hp_config: Optional[HyperparameterConfig] = None,
        device: Union[str, torch.device] = "cpu",
        accelerator: Optional[Accelerator] = None,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(index, hp_config, device, accelerator, None, name)
        assert isinstance(
            observation_space, spaces.Space
        ), "Observation space must be an instance of gymnasium.spaces.Space."
        assert isinstance(
            action_space, spaces.Space
        ), "Action space must be an instance of gymnasium.spaces.Space."

        self.observation_space = observation_space
        self.action_space = action_space
        self.zero_stage = None
        if self.accelerator is not None:
            self.zero_stage = self.accelerator.state.deepspeed_plugin.deepspeed_config[
                "zero_optimization"
            ]["stage"]
        if (
            self.zero_stage is not None
            and self.zero_stage >= 2
            and self.accelerator.is_main_process
        ):
            warnings.warn(
                "Zero stage 3 support is nascent and has not been thoroughly tested. It may be unstable or subject to change. We recommend caution in production environments."
            )

        seed = 42
        if self.accelerator is not None:
            if accelerator.is_main_process:
                seed = np.random.randint(0, 2**32 - 1)
            if accelerator.num_processes > 1:
                seed = broadcast_object_list([seed], from_process=0)[0]
        self.rng = np.random.RandomState(seed)

    def preprocess_observation(self, observation: ObservationType) -> TorchObsType:
        """Dummy preprocesses observations for forward pass through neural network.

        :param observations: Observations of environment
        :type observations: numpy.ndarray[float] or dict[str, numpy.ndarray[float]]

        :return: Preprocessed observations
        :rtype: torch.Tensor[float] or dict[str, torch.Tensor[float]] or Tuple[torch.Tensor[float], ...]
        """
        return observation

    def save_checkpoint(self, path: str) -> None:
        """
        Override the save_checkpoint method to provide guidance on the correct method to use.
        :param path: Location to save checkpoint at
        :type path: string
        """
        raise NotImplementedError(
            "The save_checkpoint method is not supported for this algorithm class. "
            "Please use agent.actor.save_pretrained(checkpoint_path) instead."
        )

    def load_checkpoint(self, path: str) -> None:
        """
        Override the load_checkpoint method to provide guidance on the correct method to use.

        :param path: Location to load checkpoint from
        :type path: string
        """
        raise NotImplementedError(
            "The load_checkpoint method is not supported for this algorithm class."
            """
            To load a saved LLM, please load the model as follows, and then re-instantiate the GRPO
            class.

            base_model = AutoModelForCausalLM.from_pretrained(
                "Qwen/Qwen2.5-3B",
                torch_dtype=torch.bfloat16,
                device_map="auto"
            )
            tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-3B")
            model = PeftModel.from_pretrained(base_model, "/path/to/adapter/folder")
            """
        )

    def _save_distributed_actor(self, path: str) -> None:
        """
        Override the save_checkpoint method to provide guidance on the correct method to use.

        :param path: Output directory to save the checkpoint at
        :type path: str
        """
        if self.accelerator is not None:
            os.makedirs(path, exist_ok=True)
            self.actor.save_checkpoint(path, tag="checkpoint")
        else:
            warnings.warn(
                "Distributed actor save not supported for non-distributed training."
            )

    def _load_distributed_actor(self, path: str) -> None:
        """
        Override the load_checkpoint method to provide guidance on the correct method to use.

        :param path: Output directory to load the checkpoint from
        :type path: str
        """
        if self.accelerator is not None:
            deepspeed_dirs = sorted(glob.glob(f"{path}/checkpoint"))
            try:
                assert len(deepspeed_dirs) > 0
                load_path, _ = self.actor.load_checkpoint(
                    path,
                    tag="checkpoint",
                    load_module_strict=False,
                    load_optimizer_states=True,
                    load_lr_scheduler_states=True,
                )
                if load_path is None:
                    raise ValueError(
                        f"Deepspeed failed to resume from checkpoint {path}"
                    )

            except Exception as e:
                raise ValueError(
                    f"Deepspeed failed to resume from checkpoint {path}"
                ) from e
        else:
            warnings.warn(
                "Distributed actor load not supported for non-distributed training."
            )

    @classmethod
    def load(
        cls,
        path: str,
        device: DeviceType = "cpu",
        accelerator: Optional[Accelerator] = None,
    ) -> None:
        raise NotImplementedError(
            "The load class method is not supported for this algorithm class."
            """
            To load a saved LLM, please load the model as follows, and then re-instantiate the GRPO
            class, using the pre-trained model.

            base_model = AutoModelForCausalLM.from_pretrained(
                "Qwen/Qwen2.5-3B",
                torch_dtype=torch.bfloat16,
                device_map="auto"
            )
            tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen2.5-3B")
            model = PeftModel.from_pretrained(base_model, "/path/to/adapter/folder")
            """
        )

    def wrap_models(self):
        """Wrap the models in the accelerator"""
        if self.accelerator is not None:
            self.actor, self.optimizer, self.lr_scheduler = self.accelerator.prepare(
                self.actor, self.optimizer.optimizer, self.lr_scheduler
            )
            deepspeed_plugin = self.accelerator.state.deepspeed_plugin
            config_kwargs = copy.deepcopy(deepspeed_plugin.deepspeed_config)
            self.reference_actor = deepspeed.init_inference(
                self.reference_actor,
                tensor_parallel={"tp_size": self.accelerator.num_processes},
                dtype=(
                    torch.bfloat16
                    if config_kwargs.get("bf16", {}).get("enabled", False)
                    else (
                        torch.float16
                        if config_kwargs.get("fp16", {}).get("enabled", False)
                        else torch.float32
                    )
                ),
                checkpoint=None,
                replace_with_kernel_inject=True,
                zero=DeepSpeedZeroConfig(**config_kwargs["zero_optimization"]),
            )
            self.reference_actor.eval()

    def clean_up(self) -> None:
        """Clean up the algorithm."""

        if self.accelerator is not None:
            self.actor, self.reference_actor, self.optimizer, self.lr_scheduler = (
                self.accelerator.free_memory(
                    self.actor, self.reference_actor, self.optimizer, self.lr_scheduler
                )
            )
            self.accelerator.wait_for_everyone()
        else:
            self.actor, self.reference_actor, self.optimizer, self.lr_scheduler = (
                None,
                None,
                None,
                None,
            )
        gc.collect()
        torch.cuda.empty_cache()

    def clone(self, index: Optional[int] = None, wrap: bool = True):
        """Creates a clone of the algorithm.

        :param index: The index of the clone, defaults to None
        :type index: Optional[int], optional
        :param wrap: If True, wrap the models in the clone with the accelerator, defaults to False
        :type wrap: bool, optional

        :return: A clone of the algorithm
        :rtype: EvolvableAlgorithm
        """
        with tempfile.TemporaryDirectory() as temp_dir:

            # We need to use the same temp_dir for all processes, so we broadcast the temp_dir from the main process
            if self.accelerator is not None and self.accelerator.num_processes > 1:
                temp_dir = broadcast_object_list([temp_dir], from_process=0)[0]

            if self.zero_stage is not None and self.zero_stage >= 2:
                self.accelerator.wait_for_everyone()
                self._save_distributed_actor(f"{temp_dir}/agent_{self.index}")
                self.accelerator.wait_for_everyone()

            input_args = EvolvableAlgorithm.inspect_attributes(
                self, input_args_only=True
            )
            input_args["clone"] = True
            input_args["wrap"] = False

            # extract base model and peft config
            actor = (
                self.accelerator.unwrap_model(self.actor)
                if self.accelerator is not None
                else self.actor
            )

            cloned_actor = clone_llm(
                actor, load_state_dict=(self.zero_stage is None or self.zero_stage < 2)
            )

            actor = None  # De-reference the actor
            input_args["actor_network"] = cloned_actor
            input_args["accelerator"] = (
                Accelerator() if self.accelerator is not None else None
            )

            clone = type(self)(**input_args)

            # Maybe this needs to change -> should really init deepspeed engine after state dict is loaded
            clone.reference_actor.load_state_dict(self.reference_actor_state_dict)
            clone.reference_actor.eval()
            clone.mutation_hook()

            # Clone attributes
            accelerator = clone.accelerator
            lr_scheduler = clone.lr_scheduler
            cloned_lr_scheduler = clone.lr_scheduler
            original_lr_scheduler = self.lr_scheduler
            clone.lr_scheduler = None
            self.lr_scheduler = None
            clone = EvolvableAlgorithm.copy_attributes(self, clone)
            clone.accelerator = accelerator
            clone.lr_scheduler = lr_scheduler
            clone.lr_scheduler = cloned_lr_scheduler
            self.lr_scheduler = original_lr_scheduler

            if self.accelerator is None:
                clone.optimizer.optimizer.load_state_dict(
                    self.optimizer.optimizer.state_dict()
                )
                if self.lr_scheduler is not None:
                    clone.lr_scheduler.load_state_dict(self.lr_scheduler.state_dict())

            # Set the index
            if index is not None:
                clone.index = index

            clone.wrap_models()
            for param, cloned_param in zip(
                self.reference_actor.parameters(), clone.reference_actor.parameters()
            ):
                cloned_param.data = param.data

            if self.zero_stage is not None and self.zero_stage >= 2:
                clone.accelerator.wait_for_everyone()
                clone._load_distributed_actor(f"{temp_dir}/agent_{self.index}")
                clone.accelerator.wait_for_everyone()
            else:
                if self.accelerator is not None:
                    self.accelerator.wait_for_everyone()
        return clone

    @staticmethod
    def update_lr(optimizer: DeepSpeedOptimizerWrapper, lr: float) -> None:
        """Update the learning rate of the optimizer

        :param optimizer: Optimizer
        :type optimizer: Optimizer
        """
        for param_group in optimizer.param_groups:
            param_group["lr"] = lr



import warnings
from contextlib import contextmanager
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple

import gymnasium as gym
import torch
from accelerate import Accelerator
from torch.utils.data import DataLoader
from transformers import AutoTokenizer
from transformers.tokenization_utils_base import BatchEncoding

REASONING_SYSTEM_PROMPT = (
    "A conversation between User and Assistant. The user asks a question, and the Assistant solves it. The assistant "
    "first thinks about the reasoning process in the mind and then provides the user with the answer. The reasoning "
    "process and answer are enclosed within <think> </think> and <answer> </answer> tags, respectively, i.e., "
    "<think> reasoning process here </think><answer> answer here </answer>"
)


class HuggingFaceGym(gym.Env):
    """Class to convert HuggingFace datasets into Gymnasium style environment.

    :param dataset_name: Dataset name to be loaded from HuggingFace datasets.
    :type dataset_name: str
    :param tokenizer: Tokenizer to be used for encoding and decoding the prompts.
    :type tokenizer: AutoTokenizer
    :param reward_fn: Reward function for evaluating completions.
    :type reward_fn: Callable[..., float]
    :param data_batch_size_per_gpu: DataLoader batch size, defaults to 8
    :type data_batch_size_per_gpu: int, optional
    :param accelerator: Accelerator to be used for training, defaults to None
    :type accelerator: Accelerator, optional
    """

    def __init__(
        self,
        train_dataset: str,
        test_dataset,
        tokenizer: AutoTokenizer,
        reward_fn: Callable[[str, str, str], float],
        apply_chat_template_fn: Callable[[str, str, AutoTokenizer], BatchEncoding],
        data_batch_size_per_gpu: int = 8,
        custom_collate_fn: Callable = None,
        accelerator: Optional[Accelerator] = None,
    ) -> None:
        assert {"question", "answer"}.issubset(
            set(train_dataset.features.keys())
        ), "Train dataset must contain 'question' and 'answer' features."
        assert {"question", "answer"}.issubset(
            set(test_dataset.features.keys())
        ), "Train dataset must contain 'question' and 'answer' features."
        self.name = train_dataset.info.dataset_name
        self.reward_fn = reward_fn
        self.tokenizer = tokenizer
        self.data_batch_size_per_gpu = data_batch_size_per_gpu
        dataloader_kwargs = (
            {} if custom_collate_fn is None else {"collate_fn": custom_collate_fn}
        )
        self.train_dataloader = DataLoader(
            train_dataset,
            batch_size=data_batch_size_per_gpu,
            shuffle=True,
            **dataloader_kwargs,
        )
        self.test_dataloader = DataLoader(
            test_dataset,
            batch_size=data_batch_size_per_gpu,
            shuffle=False,
            **dataloader_kwargs,
        )
        self.accelerator = accelerator
        if self.accelerator is not None:
            self.train_dataloader = self.accelerator.prepare(self.train_dataloader)
            self.test_dataloader = self.accelerator.prepare(self.test_dataloader)
        self.train_dataloader_iter = iter(self.train_dataloader)
        self.test_dataloader_iter = iter(self.test_dataloader)
        self.apply_chat_template_fn = apply_chat_template_fn
        self.dataloader = self.train_dataloader_iter
        self.reset_called = False
        self.observation_space = gym.spaces.Box(low=0, high=tokenizer.vocab_size - 1)
        self.action_space = gym.spaces.Box(
            low=0,
            high=tokenizer.vocab_size - 1,
        )
        self.eval_mode = False

    def step(
        self, completions: torch.Tensor
    ) -> Tuple[List[BatchEncoding], torch.Tensor]:
        """Take a step in the HuggingFaceGym environment, calculate rewards from completions generated from previous prompt and provide new batch
        of prompts.

        :param completions: Completion IDs generated by the agent.
        :type completions: torch.Tensor
        :return: New tokenized prompts and an information dictionary.
        :rtype: Tuple[List[BatchEncoding], torch.Tensor]
        """
        self.reset_called = False
        rewards = self._decode_and_evaluate(completions)
        new_tokenized_prompts = self._get_next_batch()
        self.last_tokenized_prompts = new_tokenized_prompts
        return new_tokenized_prompts, rewards

    def reset(
        self, reset_dataloaders: bool = False
    ) -> Tuple[List[BatchEncoding], Dict[str, Any]]:
        if reset_dataloaders:
            self._reset_dataloaders()
        if self.reset_called:
            warnings.warn(
                "env.reset() called more than once sequentially, it should typically follow with env.step()."
            )
        self.reset_called = True
        new_tokenized_prompts = self._get_next_batch()
        self.last_tokenized_prompts = new_tokenized_prompts
        return new_tokenized_prompts

    def _decode_and_evaluate(self, completions: List[torch.Tensor]) -> torch.Tensor:
        # This is for a batch of completions (prompt_batch x group_size), List of tensors of length batch size, each tensor is a group of answers
        total_rewards = []
        if self.eval_mode:
            decoded_completions = []
        for idx, (group_completion, answer, question) in enumerate(
            zip(completions, self.answers, self.questions)
        ):  # Vectorize this in the future
            decoded_group_completion = self.tokenizer.batch_decode(
                group_completion[
                    :, self.last_tokenized_prompts[idx]["input_ids"].shape[1] :
                ],
                skip_special_tokens=True,
            )
            if self.eval_mode:
                decoded_completions.append(decoded_group_completion)
            rewards = [
                self.reward_fn(completion, answer, question)
                for completion in decoded_group_completion
            ]
            total_rewards.append(rewards)
        return torch.tensor(total_rewards)

    def _get_next_batch(self) -> List[BatchEncoding]:
        batch = next(self.dataloader)
        self.questions = batch["question"]
        self.answers = batch["answer"]
        tokenized_prompts = [
            self.apply_chat_template_fn(question, answer, self.tokenizer)
            for question, answer in zip(self.questions, self.answers)
        ]
        return tokenized_prompts

    @contextmanager
    def eval(self) -> Generator[None, None, None]:
        self.dataloader = self.test_dataloader_iter
        self.eval_mode = True
        try:
            yield
        finally:
            self.dataloader = self.train_dataloader_iter
            self.eval_mode = False

    def __len__(self):
        if self.eval_mode:
            return len(self.test_dataloader.dataset)
        return len(self.train_dataloader.dataset)

    def _reset_dataloaders(self):
        self.train_dataloader_iter = iter(self.train_dataloader)
        self.test_dataloader_iter = iter(self.test_dataloader)



import copy
import warnings
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, TypeVar, Union

import fastrand
import numpy as np
import torch
import torch.nn as nn
from accelerate import Accelerator
from accelerate.utils.deepspeed import DeepSpeedOptimizerWrapper
from numpy.random import Generator
from torch._dynamo.eval_frame import OptimizedModule

from agilerl.algorithms.core import EvolvableAlgorithm, LLMAlgorithm
from agilerl.algorithms.core.wrappers import OptimizerWrapper
from agilerl.algorithms.neural_ts_bandit import NeuralTS
from agilerl.algorithms.neural_ucb_bandit import NeuralUCB
from agilerl.modules.base import EvolvableModule
from agilerl.protocols import OptimizerConfig
from agilerl.utils.algo_utils import remove_compile_prefix

NetworkConfig = Dict[str, str]
NetworkList = List[NetworkConfig]
SelfEvolvableAlgorithm = TypeVar("T", bound=EvolvableAlgorithm)
MutationMethod = Callable[[SelfEvolvableAlgorithm], SelfEvolvableAlgorithm]
AlgoConfig = Dict[str, Union[NetworkConfig, NetworkList]]
PopulationType = Iterable[SelfEvolvableAlgorithm]
ModuleType = Union[OptimizedModule, EvolvableModule]
OffspringType = Union[List[EvolvableModule], EvolvableModule]
MutationReturnType = Union[Dict[str, Any], List[Dict[str, Any]]]
BanditAlgorithm = Union[NeuralUCB, NeuralTS]


def set_global_seed(seed: Optional[int]) -> None:
    """Set the global seed for random number generators.

    :param seed: Random seed for repeatability
    :type seed: int
    """
    if seed is None:
        return

    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
    fastrand.pcg32_seed(seed)


def get_architecture_mut_method(
    eval: OffspringType, new_layer_prob: float, rng: Generator
) -> str:
    """Get the mutation method and its return type of the individual.

    :param individual: The individual to inspect
    :type individual: EvolvableAlgorithm
    :param new_layer_prob: Relative probability of new layer mutation (type of architecture mutation)
    :type new_layer_prob: float
    :param rng: Random number generator
    :type rng: Generator
    :return: The mutation methods name
    :rtype: str
    """
    # All of the offsprings should be the same EvolvableModule type, so we can
    # just sample the mutation method from the first offspring
    if isinstance(eval, list):
        assert all(
            isinstance(offspring, eval[0].__class__) for offspring in eval
        ), "All offspring should be of the same type."

        # NOTE: For multi-agent settings we apply the same architecture mutations to
        # all agents. However, depending on use-case it might be beneficial to have agents
        # with different architectures; please raise an issue if you would like this!
        eval = eval[0]

    return eval.sample_mutation_method(new_layer_prob, rng)


def get_offspring_eval_modules(
    individual: SelfEvolvableAlgorithm,
) -> Tuple[Dict[str, OffspringType], ...]:
    """Get the offsprings of all of the evaluation modules in the individual.

    :param individual: The individual to inspect
    :type individual: EvolvableAlgorithm

    :return: The offspring evaluation modules
    :rtype: Dict[str, OffspringType]
    """
    registry = individual.registry

    offspring_modules = {}
    offspring_policy = {}
    for group in registry.groups:
        eval_module: OffspringType = getattr(individual, group.eval)

        # Clone the offspring prior to applying mutations
        offspring = (
            [mod.clone() for mod in eval_module]
            if isinstance(eval_module, list)
            else eval_module.clone()
        )

        if group.policy:
            offspring_policy[group.eval] = offspring
        else:
            offspring_modules[group.eval] = offspring

    return offspring_policy, offspring_modules


def get_exp_layer(offspring: EvolvableModule) -> nn.Module:
    """Get the output layer of different types of offsprings for bandit algorithms.
    Returns None if algorithm is not a bandit algorithm.

    :param offspring: The offspring to inspect
    :type offspring: EvolvableModule

    :return: The output layer of the offspring
    :rtype: nn.Module
    """
    if isinstance(offspring, EvolvableModule):
        exp_layer = offspring.get_output_dense()
    else:
        raise ValueError(
            f"Bandit algorithm architecture {type(offspring)} not supported."
        )

    return exp_layer


class Mutations:
    """The Mutations class for evolutionary hyperparameter optimization.

    :param no_mutation: Relative probability of no mutation
    :type no_mutation: float
    :param architecture: Relative probability of architecture mutation
    :type architecture: float
    :param new_layer_prob: Relative probability of new layer mutation (type of architecture mutation)
    :type new_layer_prob: float
    :param parameters: Relative probability of network parameters mutation
    :type parameters: float
    :param activation: Relative probability of activation layer mutation
    :type activation: float
    :param rl_hp: Relative probability of learning hyperparameter mutation
    :type rl_hp: float
    :param rl_hp_selection: Learning hyperparameter mutations to choose from
    :type rl_hp_selection: list[str]
    :param mutation_sd: Mutation strength
    :type mutation_sd: float
    :param activation_selection: Activation functions to choose from, defaults to ["ReLU", "ELU", "GELU"]
    :type activation_selection: list[str], optional
    :param mutate_elite: Mutate elite member of population, defaults to True
    :type mutate_elite: bool, optional
    :param rand_seed: Random seed for repeatability, defaults to None
    :type rand_seed: int, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    """

    def __init__(
        self,
        no_mutation: float,
        architecture: float,
        new_layer_prob: float,
        parameters: float,
        activation: float,
        rl_hp: float,
        mutation_sd: float = 0.1,
        activation_selection: List[str] = ["ReLU", "ELU", "GELU"],
        mutate_elite: bool = True,
        rand_seed: Optional[int] = None,
        device: str = "cpu",
        accelerator: Optional[Accelerator] = None,
    ):
        assert isinstance(
            no_mutation, (float, int)
        ), "Probability of no mutation must be a float or integer."
        assert (
            no_mutation >= 0
        ), "Probability of no mutation must be greater than or equal to zero."
        assert isinstance(
            architecture, (float, int)
        ), "Probability of architecture mutation must be a float or integer."
        assert (
            architecture >= 0
        ), "Probability of architecture mutation must be greater than or equal to zero."
        assert isinstance(
            new_layer_prob, (float, int)
        ), "Probability of new layer architecture mutation must be a float or integer."
        assert (
            1 >= new_layer_prob >= 0
        ), "Probability of new layer architecture mutation must be between zero and one (inclusive)."
        assert isinstance(
            parameters, (float, int)
        ), "Probability of parameters mutation must be a float or integer."
        assert (
            parameters >= 0
        ), "Probability of parameters mutation must be greater than or equal to zero."
        assert isinstance(
            activation, (float, int)
        ), "Probability of activation mutation must be a float or integer."
        assert (
            activation >= 0
        ), "Probability of activation mutation must be greater than or equal to zero."
        assert isinstance(
            rl_hp, (float, int)
        ), "Probability of reinforcement learning hyperparameter mutation must be a float or integer."
        assert (
            rl_hp >= 0
        ), "Probability of reinforcement learning hyperparameter mutation must be greater than or equal to zero."
        assert (
            mutation_sd >= 0
        ), "Mutation strength must be greater than or equal to zero."
        assert isinstance(
            mutation_sd, (float, int)
        ), "Mutation strength must be a float or integer."
        assert isinstance(
            mutate_elite, bool
        ), "Mutate elite must be boolean value True or False."
        assert (
            isinstance(rand_seed, int) or rand_seed is None
        ), "Random seed must be an integer or None."
        if isinstance(rand_seed, int):
            assert rand_seed >= 0, "Random seed must be greater than or equal to zero."

        # Random seed for repeatability
        set_global_seed(rand_seed)
        self.rng = np.random.default_rng(rand_seed)

        # Relative probabilities of mutation
        self.no_mut = no_mutation  # No mutation
        self.architecture_mut = architecture  # Architecture mutation

        # New layer mutation (type of architecture mutation)
        self.new_layer_prob = new_layer_prob
        self.parameters_mut = parameters  # Network parameters mutation
        self.activation_mut = activation  # Activation layer mutation
        self.rl_hp_mut = rl_hp  # Learning HP mutation
        self.activation_selection = activation_selection  # Learning HPs to choose from
        self.mutation_sd = mutation_sd  # Mutation strength
        self.mutate_elite = mutate_elite
        self.device = device
        self.accelerator = accelerator

        self.pretraining_mut_options, self.pretraining_mut_proba = (
            self.get_mutations_options(pretraining=True)
        )
        self.mut_options, self.mut_proba = self.get_mutations_options()

    def get_mutations_options(
        self, pretraining: bool = False
    ) -> Tuple[List[Callable], List[float]]:
        """Get the mutation options and probabilities for the given mutation
        configuration.

        :param pretraining: Boolean flag indicating if the mutation is before the training loop
        :type pretraining: bool
        :return: Mutation functions and their respective relative probabilities
        :rtype: Tuple[List[Callable], List[float]]
        """
        # Create lists of possible mutation functions and their
        # respective relative probabilities
        mutation_options = [
            (self.no_mutation, self.no_mut),
            (self.architecture_mutate, self.architecture_mut),
            (self.parameter_mutation, self.parameters_mut),
            (self.activation_mutation, self.activation_mut),
            (self.rl_hyperparam_mutation, self.rl_hp_mut),
        ]

        if pretraining:
            mutation_options[0] = (self.no_mutation, 0)

        mutation_options = [(func, prob) for func, prob in mutation_options if prob > 0]

        # This will really only happen when pretraining is True and user has set
        # all mutation probabilities to zero, hence we apply no mutation
        if len(mutation_options) == 0:
            mutation_options = [(self.no_mutation, 1)]

        mutation_funcs, mutation_proba = zip(*mutation_options)
        mutation_proba = np.array(mutation_proba) / np.sum(mutation_proba)
        return mutation_funcs, mutation_proba

    def no_mutation(self, individual: SelfEvolvableAlgorithm):
        """Returns individual from population without mutation.

        :param individual: Individual agent from population
        :type individual: object
        """
        individual.mut = "None"  # No mutation
        return individual

    def to_device(self, offsprings: OffspringType) -> OffspringType:
        """Move offspring to device.

        :param offsprings: The offspring to move to device
        :type offsprings: OffspringType
        """
        if isinstance(offsprings, list):
            return [offspring.to(self.device) for offspring in offsprings]
        else:
            return offsprings.to(self.device)

    def to_device_and_set_individual(
        self, individual: SelfEvolvableAlgorithm, name: str, networks: OffspringType
    ) -> None:
        """Moves networks to the device and assigns them back to the individual.

        :param individual: The individual to assign the networks to
        :type individual: EvolvableAlgorithm
        :param name: The name of the attribute to assign the networks to
        :type name: str
        :param networks: The networks to move to the device
        :type networks: OffspringType
        """
        if self.accelerator is None:
            setattr(individual, name, self.to_device(networks))
        else:
            setattr(individual, name, networks)

    def reinit_module(
        self, module: EvolvableModule, init_dict: Dict[str, Any]
    ) -> EvolvableModule:
        """Reinitialize the module with the given initialization dictionary.

        :param module: The module to reinitialize
        :type module: EvolvableModule

        :param init_dict: The initialization dictionary
        :type init_dict: Dict[str, Any]
        """
        if isinstance(module, torch._dynamo.eval_frame.OptimizedModule):
            module_cls = type(module._orig_mod)
        else:
            module_cls = type(module)

        return module_cls(**init_dict)

    def reinit_from_mutated(
        self, offspring: OffspringType, remove_compile_prefix: bool = False
    ) -> OffspringType:
        """Reinitialize the mutated offspring with their state dictionary.

        :param offspring: The offspring to reinitialize
        :type offspring: OffspringType

        :return: The reinitialized offspring
        :rtype: OffspringType
        """
        if isinstance(offspring, list):
            ind_shared = [
                self.reinit_module(offspring, offspring.init_dict)
                for offspring in offspring
            ]

            # Load eval state dicts into shared networks
            state_dicts = [offspring.state_dict() for offspring in offspring]
            self.load_state_dicts(ind_shared, state_dicts, remove_compile_prefix)
        else:
            ind_shared = self.reinit_module(offspring, offspring.init_dict)
            ind_shared.load_state_dict(offspring.state_dict(), strict=False)

        return ind_shared

    def load_state_dicts(
        self,
        modules: List[ModuleType],
        state_dicts: List[Dict[str, Any]],
        remove_prefix: bool = False,
    ) -> None:
        """Load the state dictionary into the module.

        :param module: The module to load the state dictionary into
        :type module: ModuleType
        :param state_dict: The state dictionary to load
        :type state_dict: Dict[str, Any]
        :param remove_prefix: Boolean flag indicating if the compile prefix should be removed, defaults to False
        :type remove_prefix: bool, optional
        """
        for module, state_dict in zip(modules, state_dicts):
            state_dict = (
                remove_compile_prefix(state_dict) if remove_prefix else state_dict
            )
            module.load_state_dict(state_dict, strict=False)

    def compile_modules(self, modules: OffspringType, compiler: str) -> OffspringType:
        """Compile the modules using the given compiler.

        :param modules: The modules to compile
        :type modules: List[ModuleType]

        :param compiler: The compiler to use
        :type compiler: Optional[str]
        """
        single_offspring = not isinstance(modules, list)
        if single_offspring:
            modules = [modules]

        # Compile modules
        compiled_modules = []
        for module in modules:
            if not isinstance(module, torch._dynamo.eval_frame.OptimizedModule):
                compiled_modules.append(torch.compile(module, mode=compiler))
            else:
                compiled_modules.append(module)

        return compiled_modules if not single_offspring else compiled_modules[0]

    def mutation(
        self, population: PopulationType, pre_training_mut: bool = False
    ) -> PopulationType:
        """Returns mutated population.

        :param population: Population of agents
        :type population: list[EvolvableAlgorithm]
        :param pre_training_mut: Boolean flag indicating if the mutation is before the training loop
        :type pre_training_mut: bool, optional
        """
        # Create lists of possible mutation functions and their respective relative probabilities
        mutation_options = (
            self.pretraining_mut_options if pre_training_mut else self.mut_options
        )
        mutation_proba = (
            self.pretraining_mut_proba if pre_training_mut else self.mut_proba
        )

        # Randomly choose mutation for each agent in population from options with
        # relative probabilities
        mutation_choice: List[MutationMethod] = self.rng.choice(
            mutation_options, len(population), p=mutation_proba
        )

        # If not mutating elite member of population (first in list from tournament selection),
        # set this as the first mutation choice
        if not self.mutate_elite:
            mutation_choice[0] = self.no_mutation

        mutated_population = []
        for mutation, individual in zip(mutation_choice, population):
            individual: SelfEvolvableAlgorithm = individual
            registry = individual.registry

            # Call mutation function for each individual
            individual = mutation(individual)

            # Recompile modules if applicable
            compiled_individual = False
            if hasattr(individual, "torch_compiler") and individual.torch_compiler:
                compiled_individual = True
                individual.recompile()

            # Reinitiliaze shared networks to mutated evaluation networks
            for net_group in registry.groups:
                if net_group.shared is not None:
                    for shared_name in net_group.shared:
                        eval_offspring: OffspringType = getattr(
                            individual, net_group.eval
                        )

                        # Reinitialize shared with frozen weights due to
                        # potential mutation in architecture
                        ind_shared = self.reinit_from_mutated(
                            eval_offspring, remove_compile_prefix=compiled_individual
                        )

                        if self.accelerator is None:
                            ind_shared = self.to_device(ind_shared)

                        # Compile modules if necessary
                        if (
                            hasattr(individual, "torch_compiler")
                            and individual.torch_compiler
                        ):
                            ind_shared = self.compile_modules(
                                ind_shared, individual.torch_compiler
                            )

                        setattr(individual, shared_name, ind_shared)

            # Call hooks specified by user
            individual.mutation_hook()

            mutated_population.append(individual)

        return mutated_population

    def reinit_opt(
        self,
        individual: SelfEvolvableAlgorithm,
        optimizer: Optional[OptimizerConfig] = None,
    ) -> None:
        """Reinitialize the optimizers of an individual.

        :param individual: The individual to reinitialize the optimizers for
        :type individual: EvolvableAlgorithm
        """

        def _reinit_individual(config: OptimizerConfig) -> None:
            opt: Union[OptimizerWrapper, DeepSpeedOptimizerWrapper] = getattr(
                individual, config.name
            )
            optimizer = opt.optimizer

            # Multiple optimizers in a single attribute (i.e. multi-agent)
            # or one module optimized by a single optimizer
            if isinstance(opt, DeepSpeedOptimizerWrapper):
                LLMAlgorithm.update_lr(opt, individual.lr)
            else:
                if isinstance(optimizer, list) or len(opt.network_names) == 1:
                    opt_nets = getattr(individual, opt.network_names[0])

                # Multiple modules optimized by a single optimizer (e.g. PPO)
                else:
                    opt_nets = [getattr(individual, net) for net in opt.network_names]

                # Reinitialize optimizer with mutated nets
                offspring_opt = OptimizerWrapper(
                    optimizer_cls=config.get_optimizer_cls(),
                    networks=opt_nets,
                    lr=getattr(individual, opt.lr_name),
                    optimizer_kwargs=opt.optimizer_kwargs,
                    network_names=opt.network_names,
                    lr_name=opt.lr_name,
                    multiagent=opt.multiagent,
                )

                setattr(individual, config.name, offspring_opt)

        if optimizer is not None:
            _reinit_individual(optimizer)
        else:
            optimizer_configs = individual.registry.optimizers
            for opt_config in optimizer_configs:
                _reinit_individual(opt_config)

    def rl_hyperparam_mutation(
        self, individual: SelfEvolvableAlgorithm
    ) -> SelfEvolvableAlgorithm:
        """Returns individual from population with RL hyperparameter mutation.

        :param individual: Individual agent from population
        :type individual: object
        """
        # Randomly sample hyperparameter to mutate from the passed configuration
        hp_config = individual.registry.hp_config
        if not hp_config:
            individual.mut = "None"
            return individual

        mutate_attr, mutate_param = hp_config.sample()

        if mutate_param.value is None:
            mutate_param.value = getattr(individual, mutate_attr)

        # Randomly grow or shrink hyperparameters by specified factors
        new_value = mutate_param.mutate()

        setattr(individual, mutate_attr, new_value)

        # Need to reinitialize respective optimizer if mutated learning rate
        if mutate_attr in individual.get_lr_names():
            optimizer_configs = individual.registry.optimizers
            to_reinit = [
                opt_config
                for opt_config in optimizer_configs
                if mutate_attr == opt_config.lr
            ][0]
            self.reinit_opt(
                individual, optimizer=to_reinit
            )  # Reinitialise optimizer if new learning rate

        individual.mut = mutate_attr

        return individual

    # TODO: Activation mutations should really be integrated as architecture mutations
    def activation_mutation(
        self, individual: SelfEvolvableAlgorithm
    ) -> SelfEvolvableAlgorithm:
        """Returns individual from population with activation layer mutation.

        :param individual: Individual agent from population
        :type individual: EvolvableAlgorithm
        """
        # Needs to stay constant for policy gradient methods
        # NOTE: Could set up an algorithm registry to make algo checks more robust
        # OR perform activation mutations within evolvable modules directly and disable
        # on an algorithm basis
        if individual.algo in ["PPO", "DDPG", "TD3", "IPPO", "MADDPG", "MATD3", "GRPO"]:
            warnings.warn(
                f"Activation mutations are not supported for {individual.algo}."
            )
            individual.mut = "None"
            return individual

        # Mutate network activation layer
        registry = individual.registry
        no_activation = False
        for network_group in registry.groups:
            eval_module: OffspringType = getattr(individual, network_group.eval)
            if isinstance(eval_module, list):
                # TODO: Will need to modify when making multi-agent support more robust
                # to different type sof settings (i.e. different observation spaces and thus
                # network architectures for different agents)
                if eval_module[0].activation is None:
                    no_activation = True

                eval_module = [self._permutate_activation(mod) for mod in eval_module]
            else:
                if eval_module.activation is None:
                    no_activation = True

                eval_module = self._permutate_activation(eval_module)

            if no_activation:
                warnings.warn(
                    "Found no activation mutation capabilities. We advise setting the probability to "
                    "0.0 to disable activation mutations."
                )
                break
            if self.accelerator is None:
                eval_module = self.to_device(eval_module)

            if isinstance(individual, (NeuralTS, NeuralUCB)):
                individual.exp_layer = get_exp_layer(eval_module)

            setattr(individual, network_group.eval, eval_module)

        self.reinit_opt(individual)  # Reinitialise optimizer
        individual.mut = "act" if not no_activation else "None"
        return individual

    def _permutate_activation(self, network: EvolvableModule) -> EvolvableModule:
        """Permutate the activation layer of the network.

        :param network: The network to permutate the activation layer for
        :type network: EvolvableModule

        :return: The network with permutated activation layer
        :rtype: EvolvableModule
        """
        # Function to change network activation layer
        possible_activations = copy.deepcopy(self.activation_selection)
        current_activation = network.activation

        # Remove current activation from options to ensure different new
        # activation layer
        if len(possible_activations) > 1 and current_activation in possible_activations:
            possible_activations.remove(current_activation)

        new_activation = self.rng.choice(possible_activations, size=1)[
            0
        ]  # Select new activation
        network.change_activation(
            new_activation, output=False
        )  # Change activation layer
        return network

    def parameter_mutation(
        self, individual: SelfEvolvableAlgorithm
    ) -> SelfEvolvableAlgorithm:
        """Returns individual from population with network parameters mutation.

        :param individual: Individual agent from population
        :type individual: EvolvableAlgorithm
        """
        if isinstance(individual, LLMAlgorithm):
            warnings.warn("Parameter mutations are not supported for LLM algorithms.")
            individual.mut = "None"
            return individual

        registry = individual.registry

        # We only apply parameter mutations to the evaluation policy network
        # (i.e. the network used to select actions)
        offspring_policy: OffspringType = getattr(individual, registry.policy)
        if isinstance(offspring_policy, list):
            offspring_policy = [
                self.classic_parameter_mutation(mod) for mod in offspring_policy
            ]
        else:
            offspring_policy = self.classic_parameter_mutation(offspring_policy)

        if self.accelerator is None:
            offspring_policy = self.to_device(offspring_policy)

        setattr(individual, registry.policy, offspring_policy)

        self.reinit_opt(individual)  # Reinitialise optimizer
        individual.mut = "param"
        return individual

    def classic_parameter_mutation(self, network: EvolvableModule) -> EvolvableModule:
        """
        Returns network with mutated weights, with a vectorized inner loop for efficiency.

        :param network: Neural network to mutate.
        :type network: EvolvableModule
        :return: Mutated network.
        :rtype: EvolvableModule
        """
        # Parameters controlling mutation strength and probabilities
        mut_strength = self.mutation_sd
        num_mutation_frac = 0.1
        super_mut_strength = 10
        super_mut_prob = 0.05
        reset_prob = super_mut_prob + 0.05
        mag_limit = 1000000

        model_params = network.state_dict()

        # Collect keys corresponding to weight matrices (ignoring normalization params)
        potential_keys = [
            key
            for key in model_params
            if "norm" not in key and len(model_params[key].shape) == 2
        ]

        # Randomly choose a subset of keys to mutate
        how_many = int(self.rng.integers(1, len(potential_keys) + 1))
        chosen_keys = self.rng.choice(potential_keys, how_many, replace=False)

        for key in chosen_keys:
            W = model_params[key]
            num_weights = W.shape[0] * W.shape[1]
            num_mutations = int(np.ceil(num_mutation_frac * num_weights))
            if num_mutations < 1:
                continue

            # Vectorized generation of random indices (for rows and columns)
            rows = self.rng.integers(0, W.shape[0], size=num_mutations)
            cols = self.rng.integers(0, W.shape[1], size=num_mutations)
            rand_vals = self.rng.uniform(0, 1, size=num_mutations)

            # Convert indices and random values to torch tensors on the same device as W
            rows_tensor = torch.tensor(rows, dtype=torch.long, device=W.device)
            cols_tensor = torch.tensor(cols, dtype=torch.long, device=W.device)
            rand_vals_tensor = torch.tensor(rand_vals, dtype=W.dtype, device=W.device)

            # Get current weight values at the selected indices
            current_vals: torch.Tensor = W[rows_tensor, cols_tensor]
            new_vals = current_vals.clone()

            # Create masks for the different mutation types
            mask_super = rand_vals_tensor < super_mut_prob
            mask_reset = (rand_vals_tensor >= super_mut_prob) & (
                rand_vals_tensor < reset_prob
            )
            mask_normal = rand_vals_tensor >= reset_prob

            # Super mutation: add noise with std proportional to the absolute current value times super_mut_strength
            if mask_super.sum() > 0:
                std_super = (super_mut_strength * current_vals[mask_super]).abs()
                noise_super = torch.normal(
                    mean=torch.zeros_like(std_super), std=std_super
                )
                new_vals[mask_super] = current_vals[mask_super] + noise_super

            # Reset mutation: completely reset the weight using N(0, 1)
            if mask_reset.sum() > 0:
                noise_reset = torch.normal(
                    mean=torch.zeros(mask_reset.sum(), device=W.device),
                    std=torch.ones(mask_reset.sum(), device=W.device),
                )
                new_vals[mask_reset] = noise_reset

            # Normal mutation: add noise with std proportional to the absolute current value times mut_strength
            if mask_normal.sum() > 0:
                std_normal = (mut_strength * current_vals[mask_normal]).abs()
                noise_normal = torch.normal(
                    mean=torch.zeros_like(std_normal), std=std_normal
                )
                new_vals[mask_normal] = current_vals[mask_normal] + noise_normal

            # Integrate regularization by clamping all mutated values at once.
            # This is equivalent to your regularize_weight function.
            new_vals = new_vals.clamp(min=-mag_limit, max=mag_limit)

            # Write the mutated, clamped values back to the weight tensor
            W[rows_tensor, cols_tensor] = new_vals
            if self.accelerator is None:
                network = network.to(self.device)

        return network

    def architecture_mutate(
        self, individual: SelfEvolvableAlgorithm
    ) -> SelfEvolvableAlgorithm:
        """Returns individual from population with network architecture mutation, which
        adds either layers or nodes to different types of network architectures.

        :param individual: Individual agent from population
        :type individual: object
        """

        if isinstance(individual, LLMAlgorithm):
            warnings.warn(
                "Architecture mutations are not supported for LLM algorithms."
            )
            individual.mut = "None"
            return individual

        # Get the offspring evaluation modules
        # We first extract and apply a mutation for the algo policy and then apply
        # the same mutation to the rest of the evaluation modules e.g. critics
        policy, offspring_evals = get_offspring_eval_modules(individual)
        policy_name, policy_offspring = list(policy.items())[0]

        sample_policy = (
            policy_offspring[0]
            if isinstance(policy_offspring, list)
            else policy_offspring
        )
        if not sample_policy.mutation_methods:
            warnings.warn(
                "No mutation methods found for the policy network. Skipping architecture mutation. "
                "We advise setting the probability of architecture mutations to zero."
            )
            individual.mut = "None"
            return individual

        # Sample mutation method from policy network
        mut_method = get_architecture_mut_method(
            policy_offspring, self.new_layer_prob, self.rng
        )

        applied_mutations, mut_dict = self._apply_arch_mutation(
            policy_offspring, mut_method
        )
        self.to_device_and_set_individual(individual, policy_name, policy_offspring)

        if isinstance(individual, (NeuralTS, NeuralUCB)):
            old_exp_layer = get_exp_layer(policy_offspring)
            self._reinit_bandit_grads(individual, policy_offspring, old_exp_layer)

        # Apply the same mutation to the rest of the evaluation modules
        for name, offsprings in offspring_evals.items():
            self._apply_arch_mutation(offsprings, applied_mutations, mut_dict)
            self.to_device_and_set_individual(individual, name, offsprings)

            # Reinitialize bandit gradients after architecture mutation
            if isinstance(individual, (NeuralTS, NeuralUCB)):
                old_exp_layer = get_exp_layer(offsprings)
                self._reinit_bandit_grads(individual, offsprings, old_exp_layer)

        individual.mutation_hook()  # Apply mutation hook

        self.reinit_opt(individual)  # Reinitialise optimizer
        individual.mut = (
            applied_mutations[0]
            if isinstance(applied_mutations, list)
            else applied_mutations
        )
        return individual

    def _apply_arch_mutation(
        self,
        networks: OffspringType,
        mut_method: Union[Optional[str], List[Optional[str]]],
        applied_mut_dict: Optional[Union[Dict[str, Any], List[Dict[str, Any]]]] = None,
    ) -> Tuple[Union[str, List[str]], MutationReturnType]:
        """Applies the mutation method to networks and returns mutation data if needed.

        :param networks: The networks to apply the mutation to
        :type networks: OffspringType
        :param mut_method: The mutation method to apply
        :type mut_method: str
        :param ret_type: The return type of the mutation method
        :type ret_type: Type
        :param mut_dict: The mutation dictionary, defaults to None
        :type mut_dict: Optional[Union[Dict[str, Any], List[Dict[str, Any]]]], optional

        :return: The mutation method name and the mutation dictionary
        :rtype: Tuple[Union[str, List[str]], MutationReturnType]
        """
        if applied_mut_dict is None:
            applied_mut_dict = (
                [{}] * len(networks) if isinstance(networks, list) else {}
            )

        mut_dict = None
        if isinstance(networks, list):
            if isinstance(mut_method, str) or mut_method is None:
                mut_method = [mut_method] * len(networks)

            mut_dict = []
            applied_muts = []
            for i, net in enumerate(networks):
                _to_apply = mut_method[i]
                if _to_apply is None:
                    mut_dict.append({})
                    applied_muts.append(None)
                    net.last_mutation_attr = None
                    net.last_mutation = None
                    continue

                mut_return = getattr(net, mut_method[i])(**applied_mut_dict[i])
                mut_dict.append(mut_return if mut_return is not None else {})
                applied_muts.append(net.last_mutation_attr)
        else:
            if mut_method is None:
                mut_dict = {}
                networks.last_mutation_attr = None
                networks.last_mutation = None
            else:
                mut_dict = getattr(networks, mut_method)(**applied_mut_dict)

            mut_dict = mut_dict if mut_dict is not None else {}
            applied_muts = networks.last_mutation_attr

        return applied_muts, mut_dict

    # TODO: This can be implemented as a mutation hook for the bandit algorithms
    def _reinit_bandit_grads(
        self,
        individual: BanditAlgorithm,
        offspring_actor: EvolvableModule,
        old_exp_layer: nn.Module,
    ) -> None:
        """Reinitialise bandit gradients after architecture mutation.

        :param individual: Individual agent from population
        :type individual: EvolvableAlgorithm
        :param offspring_actor: Offspring actor network
        :type offspring_actor: EvolvableModule
        :param old_exp_layer: Old linear layer
        :type old_exp_layer: nn.Module
        """
        if isinstance(offspring_actor, EvolvableModule):
            exp_layer = offspring_actor.get_output_dense()
        else:
            raise ValueError(
                f"Bandit algorithm architecture {type(offspring_actor)} not supported."
            )

        individual.numel = sum(
            w.numel() for w in exp_layer.parameters() if w.requires_grad
        )
        individual.theta_0 = torch.cat(
            [w.flatten() for w in exp_layer.parameters() if w.requires_grad]
        )

        # create matrix that is copy of sigma inv
        # first go through old params, figure out which to remove, then remove any difference
        # then go through new params, figure out where to add, then add zeros/lambda
        new_sigma_inv = copy.deepcopy(individual.sigma_inv).cpu().numpy()
        old_params = dict(old_exp_layer.named_parameters())
        new_params = dict(exp_layer.named_parameters())

        to_remove = []
        i = 0
        for key, param in old_exp_layer.named_parameters():
            if param.requires_grad:
                old_size = param.numel()
                if key not in new_params.keys():
                    to_remove += list(range(i, i + old_size))
                else:
                    new_size = new_params[key].numel()
                    if new_size < old_size:
                        to_remove += list(range(i + new_size, i + old_size))
                i += old_size

        to_add = []
        i = 0
        for key, param in exp_layer.named_parameters():
            if param.requires_grad:
                new_size = param.numel()
                if key in old_params.keys():
                    old_size = old_params[key].numel()
                    if new_size > old_size:
                        to_add += list(range(i + old_size, i + new_size))
                else:
                    to_add += list(range(i, i + new_size))
                i += new_size

        # Adjust indices to add after deletion
        to_remove = np.array(to_remove)
        to_add = np.array(to_add)
        to_add -= np.sum(to_add[:, np.newaxis] > to_remove, axis=1)
        to_add -= np.arange(len(to_add))

        # Remove elements corresponding to old params
        if len(to_remove) > 0:
            new_sigma_inv = np.delete(
                np.delete(new_sigma_inv, to_remove, 0), to_remove, 1
            )

        # Add new zeros corresponding to new params, make lambda down identity diagonal
        if len(to_add) > 0:
            new_sigma_inv = np.insert(
                np.insert(new_sigma_inv, to_add, 0, 0), to_add, 0, 1
            )
            for i in to_add:
                new_sigma_inv[i, i] = individual.lamb

        individual.exp_layer = exp_layer
        individual.sigma_inv = torch.from_numpy(new_sigma_inv).to(
            individual.device
            if individual.accelerator is None
            else individual.accelerator.device
        )



import copy
from collections import OrderedDict
from dataclasses import asdict
from typing import Any, Dict, Literal, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
from gymnasium import spaces

from agilerl.modules import EvolvableCNN, EvolvableLSTM, EvolvableMLP
from agilerl.modules.base import EvolvableModule, ModuleDict, MutationType, mutation
from agilerl.modules.configs import CnnNetConfig, LstmNetConfig, MlpNetConfig, NetConfig
from agilerl.typing import ArrayOrTensor, ConfigType
from agilerl.utils.evolvable_networks import (
    get_activation,
    is_box_space_ndim,
    is_image_space,
    is_vector_space,
    tuple_to_dict_space,
)

ModuleType = Union[EvolvableModule, nn.Module]
SupportedEvolvableTypes = Union[EvolvableCNN, EvolvableMLP, EvolvableLSTM]
MultiInputConfigType = Union[ConfigType, Dict[str, ConfigType]]
TupleOrDictSpace = Union[spaces.Tuple, spaces.Dict]
TupleOrDictObservation = Union[Dict[str, ArrayOrTensor], Tuple[ArrayOrTensor]]

# Default configurations for the feature extractors
DefaultCnnConfig = CnnNetConfig(
    channel_size=[16, 16],
    kernel_size=[3, 3],
    stride_size=[1, 1],
    output_activation="ReLU",
)
DefaultMlpConfig = MlpNetConfig(
    hidden_size=[64, 64],
    output_activation="ReLU",
)
DefaultLstmConfig = LstmNetConfig(
    hidden_size=64,
    num_layers=2,
    output_activation="ReLU",
)


def get_total_flatdim(observation_spaces: spaces.Dict) -> int:
    """Get the total flat dimension of the observation space.

    :param observation_space: Dictionary or Tuple space of observations.
    :type observation_space: spaces.Dict or spaces.Tuple
    :return: Total flat dimension of the observation space.
    """
    return sum([spaces.flatdim(space) for space in observation_spaces.spaces.values()])


def is_adhoc_vector_space(space: spaces.Space, recurrent: bool = False) -> bool:
    """Check if the space is a vector space.

    :param space: Input space
    :type space: spaces.Space
    :param recurrent: Whether to check if the space is a 2D Box space, defaults to False
    :type recurrent: bool, optional
    :return: True if the space is a vector space, False otherwise
    """
    return (
        is_vector_space(space)
        or (is_box_space_ndim(space, 2) and not recurrent)
        or (isinstance(space, spaces.Box) and len(space.shape) > 3)
    )


class EvolvableMultiInput(EvolvableModule):
    """Evolvable multi-input network for `Tuple` or `Dict` observation spaces. It inspects the
    observation space to determine the type of network to build for each key. It builds an
    `EvolvableCNN` for image subspaces, an `EvolvableLSTM` for time-series subspaces if specified
    through the `recurrent` flag, and a `nn.Flatten()` for other types. Vector observations are
    concatenated with the extracted features before passing through an `EvolvableMLP` to produce
    the output tensor. Optionally, users may specify an additional `EvolvableMLP` to be applied to
    the concatenated vector observations before concatenation with the extracted features.

    :param observation_space: Dictionary or Tuple space of observations.
    :type observation_space: spaces.Dict or spaces.Tuple
    :param num_outputs: Dimension of the output tensor.
    :type num_outputs: int
    :param latent_dim: Dimension of the latent space representation. Default is 16.
    :type latent_dim: int, optional
    :param vector_space_mlp: Whether to use an MLP for the vector spaces. This is done by concatenating the
        flattened observations and passing them through an `EvolvableMLP`. Default is False, whereby the
        observations are concatenated directly to the feature encodings before the final MLP.
    :type vector_space_mlp: bool, optional
    :param cnn_config: Configuration for the CNN feature extractor. Default is None.
    :type cnn_config: MultiInputConfigType, optional
    :param mlp_config: Configuration for the MLP feature extractor. Default is None.
    :type mlp_config: MultiInputConfigType, optional
    :param lstm_config: Configuration for the LSTM feature extractor. Default is None.
    :type lstm_config: MultiInputConfigType, optional
    :param init_dicts: Dictionary of initialization dictionaries for the feature extractors. Default is {}.
    :type init_dicts: Dict[str, Dict[str, Any]], optional
    :param output_activation: Activation function for the output layer. Default is None.
    :type output_activation: Optional[str], optional
    :param recurrent: Whether to use an `EvolvableLSTM` for 2D Box spaces. Otherwise, the observation
        is flattened and treated as a vector space. Default is False.
    :type recurrent: bool, optional
    :param min_latent_dim: Minimum dimension of the latent space. Default is 8.
    :type min_latent_dim: int, optional
    :param max_latent_dim: Maximum dimension of the latent space. Default is 128.
    :type max_latent_dim: int, optional
    :param device: Device to use for the network. Default is "cpu".
    :type device: str, optional
    :param name: Name of the network. Default is "multi_input".
    :type name: str, optional
    """

    feature_net: ModuleDict
    _SupportedSpaces = (spaces.Box, spaces.Discrete, spaces.MultiDiscrete)

    def __init__(
        self,
        observation_space: TupleOrDictSpace,
        num_outputs: int,
        latent_dim: int = 16,
        vector_space_mlp: bool = False,
        cnn_config: Optional[MultiInputConfigType] = None,
        mlp_config: Optional[MultiInputConfigType] = None,
        lstm_config: Optional[MultiInputConfigType] = None,
        init_dicts: Optional[MultiInputConfigType] = None,
        output_activation: Optional[str] = None,
        recurrent: bool = False,
        min_latent_dim: int = 8,
        max_latent_dim: int = 128,
        device: str = "cpu",
        name: str = "multi_input",
    ):
        super().__init__(device)

        assert num_outputs > 0, "Number of outputs must be greater than 0."
        assert latent_dim > 0, "Latent dimension must be greater than 0."
        assert isinstance(
            observation_space, (spaces.Dict, spaces.Tuple)
        ), "Observation space must be a Dict or Tuple space."
        assert (
            latent_dim <= max_latent_dim
        ), "Latent dimension must be less than or equal to max latent dimension."
        assert (
            latent_dim >= min_latent_dim
        ), "Latent dimension must be greater than or equal to min latent dimension."

        subspaces = (
            observation_space.spaces.values()
            if isinstance(observation_space, spaces.Dict)
            else observation_space.spaces
        )
        assert all(
            [isinstance(space, self._SupportedSpaces) for space in subspaces]
        ), "Observation space must contain only Box, Discrete, or MultiDiscrete spaces."

        # Convert Tuple space to Dict space for consistency
        self.is_tuple_space = False
        if isinstance(observation_space, spaces.Tuple):
            observation_space = tuple_to_dict_space(observation_space)
            self.is_tuple_space = True

        self.observation_space = observation_space
        self.num_outputs = num_outputs
        self.cnn_config = cnn_config or DefaultCnnConfig
        self.mlp_config = mlp_config or DefaultMlpConfig
        self.lstm_config = lstm_config or DefaultLstmConfig
        self._init_dicts = init_dicts or {}
        self._activation = None
        self.mlp_name = None
        self.vector_space_mlp = vector_space_mlp
        self.latent_dim = latent_dim
        self.output_activation = output_activation
        self.min_latent_dim = min_latent_dim
        self.max_latent_dim = max_latent_dim
        self.recurrent = recurrent
        self.name = name
        self.device = device

        self.vector_spaces = spaces.Dict(
            {
                key: space
                for key, space in observation_space.spaces.items()
                if is_adhoc_vector_space(space, self.recurrent)
            }
        )
        self.total_vector_dims = get_total_flatdim(self.vector_spaces)

        # Build feature extractor
        self.feature_net = self.build_feature_extractor()

        # Calculate total extracted features from non-vector spaces
        # (i.e. Box spaces with more than one dimension)
        self.extracted_features_dim = self.calc_extracted_features_dim()

        # Vector observations (i.e. 1D Box, Discrete, MultiDiscrete) are either
        # passed through an MLP or concatenated directly to the extracted features
        features_dim = self.extracted_features_dim + self.total_vector_dims * (
            1 - self.vector_space_mlp
        )
        # Final dense layer to convert feature encodings to desired num_outputs
        self.final_dense = nn.Linear(features_dim, num_outputs, device=device)
        self.output = get_activation(output_activation)

    @property
    def net_config(self) -> Dict[str, Any]:
        """Returns the configuration of the network.

        :return: Network configuration
        :rtype: Dict[str, Any]
        """
        net_config = self.init_dict.copy()
        for attr in ["observation_space", "num_outputs", "device", "name"]:
            net_config.pop(attr)
        return net_config

    @property
    def activation(self) -> str:
        """Get the activation function for the network.

        :return: Activation function
        :rtype: str
        """
        return self._activation

    @activation.setter
    def activation(self, activation: str) -> None:
        """Set the activation function for the network.

        :param activation: Activation function to use.
        :type activation: str
        """
        self._activation = activation
        for module in self.feature_net.modules().values():
            module.change_activation(activation, output=True)

    @property
    def init_dicts(self) -> Dict[str, Dict[str, Any]]:
        """Returns the initialization dictionaries for the network.

        :return: Initialization dictionaries
        :rtype: Dict[str, Dict[str, Any]]
        """
        if not hasattr(self, "feature_net"):
            return self._init_dicts

        reformatted_dicts = {}
        for key, net in self.feature_net.modules().items():
            init_dict = net.init_dict
            init_dict.pop("input_size", None)  # LSTM
            init_dict.pop("input_shape", None)  # CNN
            init_dict.pop("num_inputs", None)  # MLP
            init_dict["num_outputs"] = self.latent_dim
            reformatted_dicts[key] = init_dict

        return reformatted_dicts

    @property
    def cnn_init_dict(self) -> Dict[str, Any]:
        """Returns the initialization dictionary for the CNN."""
        return copy.deepcopy(self.cnn_config)

    @property
    def mlp_init_dict(self) -> Dict[str, Any]:
        """Returns the initialization dictionary for the MLP."""
        return copy.deepcopy(self.mlp_config)

    @property
    def lstm_init_dict(self) -> Dict[str, Any]:
        """Returns the initialization dictionary for the LSTM."""
        return copy.deepcopy(self.lstm_config)

    def init_weights_gaussian(
        self, std_coeff: float = 4, output_coeff: float = 4
    ) -> None:
        """Initialise weights of linear layers using Gaussian distribution."""
        for module in self.feature_net.modules().values():
            module.init_weights_gaussian(std_coeff=std_coeff)

        # Initialise final dense layer
        EvolvableModule.init_weights_gaussian(self.final_dense, std_coeff=output_coeff)

    def calc_extracted_features_dim(self) -> int:
        """Calculates the toal dimension of the features extracted by the evolvable
        feature extractors.

        :return: Total dimension of the extracted features.
        :rtype: int
        """
        return sum(
            [
                self.latent_dim
                for name in self.feature_net.keys()
                if name not in self.vector_spaces.keys()
            ]
        )

    def get_inner_init_dict(
        self, key: str, default: Literal["cnn", "mlp", "lstm"]
    ) -> ConfigType:
        """Returns the initialization dictionary for the specified key.

        :param key: Key of the observation space.
        :type key: str
        :param default: Default value to return if the key is not found.
        :type default: Literal["cnn", "mlp", "lstm"]
        :return: Initialization dictionary.
        :rtype: ConfigType
        """
        init_dicts = self.init_dicts
        if key in init_dicts:
            init_dict = init_dicts[key]
            init_dict["num_outputs"] = self.latent_dim
            init_dict["device"] = self.device
            return init_dict

        init_dict = {
            "cnn": self.cnn_init_dict,
            "mlp": self.mlp_init_dict,
            "lstm": self.lstm_init_dict,
        }.get(default)

        if init_dict is None:
            raise ValueError(
                "Invalid default value provided, must be 'cnn' or 'mlp' or 'lstm'."
            )
        else:
            nested_dict = init_dict.get(key)
            init_dict = (
                copy.deepcopy(nested_dict)
                if nested_dict is not None
                else copy.deepcopy(init_dict)
            )

            if isinstance(init_dict, NetConfig):
                init_dict = asdict(init_dict)

        init_dict["num_outputs"] = self.latent_dim
        init_dict["device"] = self.device
        return init_dict

    def build_feature_extractor(self) -> Dict[str, SupportedEvolvableTypes]:
        """Creates the feature extractor and final MLP networks.

        :return: Dictionary of feature extractors.
        :rtype: Dict[str, Union[EvolvableCNN, EvolvableMLP]]
        """
        # Automatically build feature extractors from subspaces
        feature_net = ModuleDict(device=self.device)
        for key, space in self.observation_space.spaces.items():
            if isinstance(space, spaces.Box) and len(space.shape) in [0, 1]:
                continue
            # EvolvableCNN for image spaces
            if is_image_space(space):
                init_dict = copy.deepcopy(self.get_inner_init_dict(key, default="cnn"))
                feature_extractor = EvolvableCNN(
                    input_shape=space.shape,
                    name=init_dict.pop("name", key),
                    **init_dict,
                )
            elif (
                isinstance(space, spaces.Box)
                and len(space.shape) == 2
                and key not in self.vector_spaces.keys()
            ):
                # EvolvableLSTM for 2D Box spaces if recurrent=True
                init_dict = copy.deepcopy(self.get_inner_init_dict(key, default="lstm"))
                feature_extractor = EvolvableLSTM(
                    input_size=space.shape[1],
                    name=init_dict.pop("name", key),
                    **init_dict,
                )
            # Flatten other observation types
            else:
                feature_extractor = nn.Flatten()

            feature_net[key] = feature_extractor

        # Optionally, use an EvolvableMLP for all concatenated vector inputs
        if self.vector_space_mlp:
            init_dict = copy.deepcopy(
                self.get_inner_init_dict("vector_mlp", default="mlp")
            )
            self.mlp_name = init_dict.pop("name", "vector_mlp")
            vector_mlp = EvolvableMLP(
                num_inputs=self.total_vector_dims, name=self.mlp_name, **init_dict
            )
            feature_net[self.mlp_name] = vector_mlp

        return feature_net

    def forward(self, x: TupleOrDictObservation) -> torch.Tensor:
        """Forward pass of the composed network. Extracts features from each observation key and concatenates
        them with the corresponding observation key if specified. The concatenated features are then passed
        through the final MLP to produce the output tensor.

        :param x: Dictionary of observations.
        :type x: Dict[str, ArrayOrTensor], Tuple[ArrayOrTensor]
        :return: Output tensor.
        :rtype: torch.Tensor
        """
        if isinstance(x, tuple):
            x = dict(zip(self.observation_space.spaces.keys(), x))

        for key, obs in x.items():
            if not isinstance(obs, torch.Tensor):
                x[key] = torch.tensor(obs, device=self.device, dtype=torch.float32)

        # Extract features from non-vector subspaces
        extracted_features = OrderedDict()
        if self.extracted_features_dim > 0:
            for key in x.keys():
                if key in self.feature_net.keys():
                    extracted_features[key] = self.feature_net[key](x[key])

        # Extract raw features from vector spaces
        vector_inputs = []
        for key, space in self.vector_spaces.items():
            _obs = (
                extracted_features.pop(key)
                if key in extracted_features.keys()
                else x[key]
            )
            if len(_obs.shape) == 1:
                dim = len(space.shape) - 1
                _obs = _obs.unsqueeze(dim)

            vector_inputs.append(_obs)

        # Concatenate vector inputs and, optionally, pass through additional EvolvableMLP
        vector_inputs = torch.cat(vector_inputs, dim=1)
        vector_features = (
            self.feature_net[self.mlp_name](vector_inputs)
            if self.vector_space_mlp
            else vector_inputs
        )

        if extracted_features:
            extracted_features = torch.cat(list(extracted_features.values()), dim=1)
        else:
            extracted_features = torch.tensor([], device=self.device)

        # Concatenate all features and pass through final MLP
        features = torch.cat([extracted_features, vector_features], dim=1)
        latent = self.final_dense(features)
        return self.output(latent)

    @mutation(MutationType.ACTIVATION)
    def change_activation(self, activation: str, output: bool = False) -> None:
        """Set the activation function for the network.

        :param activation: Activation function to use.
        :type activation: str
        :param output: Flag indicating whether to set the output activation function, defaults to False
        :type output: bool, optional

        :return: Activation function
        :rtype: str
        """
        self.activation = activation
        if output:
            self.output = get_activation(activation)

    @mutation(MutationType.NODE)
    def add_latent_node(self, numb_new_nodes: Optional[int] = None) -> Dict[str, Any]:
        """Add a latent node to the network.

        :param numb_new_nodes: Number of new nodes to add, defaults to None
        :type numb_new_nodes: int, optional

        :return: Dictionary specifying the number of nodes added.
        :rtype: Dict[str, Any]
        """
        if numb_new_nodes is None:
            numb_new_nodes = np.random.choice([8, 16, 32], 1)[0]

        if self.latent_dim + numb_new_nodes < self.max_latent_dim:
            self.latent_dim += numb_new_nodes

        return {"numb_new_nodes": numb_new_nodes}

    @mutation(MutationType.NODE)
    def remove_latent_node(
        self, numb_new_nodes: Optional[int] = None
    ) -> Dict[str, Any]:
        """Remove a latent node from the network.

        :param numb_new_nodes: Number of nodes to remove, defaults to None
        :type numb_new_nodes: int, optional

        :return: Dictionary specifying the number of nodes removed.
        :rtype: Dict[str, Any]
        """
        if numb_new_nodes is None:
            numb_new_nodes = np.random.choice([8, 16, 32], 1)[0]

        if self.latent_dim - numb_new_nodes > self.min_latent_dim:
            self.latent_dim -= numb_new_nodes

        return {"numb_new_nodes": numb_new_nodes}

    def recreate_network(self) -> None:
        """Recreates the network with the new latent dimension."""
        feature_net = self.build_feature_extractor()
        self.feature_net = EvolvableModule.preserve_parameters(
            old_net=self.feature_net, new_net=feature_net
        )

        # Calculate total extracted features dimension
        extracted_features_dim = self.calc_extracted_features_dim()
        features_dim = extracted_features_dim + self.total_vector_dims * (
            1 - self.vector_space_mlp
        )
        final_dense = nn.Linear(features_dim, self.num_outputs, device=self.device)
        self.final_dense = EvolvableModule.preserve_parameters(
            old_net=self.final_dense, new_net=final_dense
        )



from typing import Optional, Tuple, Type, Union

import torch
from gymnasium import spaces

from agilerl.modules.base import EvolvableModule
from agilerl.modules.configs import MlpNetConfig
from agilerl.networks.base import EvolvableNetwork
from agilerl.networks.distributions import EvolvableDistribution
from agilerl.typing import ArrayOrTensor, ConfigType, TorchObsType


class DeterministicActor(EvolvableNetwork):
    """Deterministic actor network for policy-gradient algorithms. Given an observation,
    it outputs the mean of the action distribution. This is useful for e.g. DDPG, SAC, TD3.

    :param observation_space: Observation space of the environment.
    :type observation_space: spaces.Space
    :param action_space: Action space of the environment
    :type action_space: Union[spaces.Box, spaces.Discrete]
    :param encoder_cls: Encoder class to use for the network. Defaults to None, whereby it is
        automatically built using an AgileRL module according the observation space.
    :type encoder_cls: Optional[Union[str, Type[EvolvableModule]]]
    :param encoder_config: Configuration of the encoder network.
    :type encoder_config: ConfigType
    :param head_config: Configuration of the network MLP head.
    :type head_config: Optional[ConfigType]
    :param clip_actions: Whether to clip the actions to the action space.
    :type clip_actions: bool
    :param min_latent_dim: Minimum dimension of the latent space representation.
    :type min_latent_dim: int
    :param max_latent_dim: Maximum dimension of the latent space representation.
    :type max_latent_dim: int
    :param n_agents: Number of agents in the environment. Defaults to None, which corresponds to
        single-agent environments.
    :type n_agents: Optional[int]
    :param latent_dim: Dimension of the latent space representation.
    :type latent_dim: int
    :param simba: Whether to use the SimBa architecture for training the network.
    :type simba: bool
    :param recurrent: Whether to use a recurrent network.
    :type recurrent: bool
    :param device: Device to use for the network.
    :type device: str
    """

    supported_spaces = (spaces.Box, spaces.Discrete)

    def __init__(
        self,
        observation_space: spaces.Space,
        action_space: Union[spaces.Box, spaces.Discrete],
        encoder_cls: Optional[Union[str, Type[EvolvableModule]]] = None,
        encoder_config: Optional[ConfigType] = None,
        head_config: Optional[ConfigType] = None,
        clip_actions: bool = True,
        min_latent_dim: int = 8,
        max_latent_dim: int = 128,
        n_agents: Optional[int] = None,
        latent_dim: int = 32,
        simba: bool = False,
        recurrent: bool = False,
        device: str = "cpu",
    ):

        super().__init__(
            observation_space,
            encoder_cls=encoder_cls,
            encoder_config=encoder_config,
            action_space=action_space,
            min_latent_dim=min_latent_dim,
            max_latent_dim=max_latent_dim,
            n_agents=n_agents,
            latent_dim=latent_dim,
            simba=simba,
            recurrent=recurrent,
            device=device,
        )

        if isinstance(action_space, spaces.Box):
            self.action_low = torch.as_tensor(action_space.low, device=self.device)
            self.action_high = torch.as_tensor(action_space.high, device=self.device)
        else:
            self.action_low = None
            self.action_high = None

        # Set output activation based on action space
        if isinstance(head_config, dict) and "output_activation" in head_config:
            output_activation = head_config["output_activation"]
        elif isinstance(action_space, spaces.Box):
            # Squash output by default if continuous action space
            output_activation = "Tanh"
        elif isinstance(action_space, spaces.Discrete):
            output_activation = "Softmax"
        else:
            output_activation = None

        self.clip_actions = clip_actions
        if head_config is None:
            head_config = MlpNetConfig(
                hidden_size=[32], output_activation=output_activation
            )
        else:
            head_config["output_activation"] = output_activation

        self.build_network_head(head_config)
        self.output_activation = head_config.get("output_activation", output_activation)

    @torch.compiler.disable
    @staticmethod
    def rescale_action(
        action: torch.Tensor,
        low: torch.Tensor,
        high: torch.Tensor,
        output_activation: Optional[str] = None,
    ) -> torch.Tensor:
        """Rescale an action to the original action space.

        :param action: Action.
        :type action: torch.Tensor
        :param low: Minimum action.
        :type low: torch.Tensor
        :param high: Maximum action.
        :type high: torch.Tensor
        :param output_activation: Output activation function.
        :type output_activation: Optional[str]
        :return: Rescaled action.
        :rtype: torch.Tensor
        """
        if output_activation in ["Tanh", "Softsign"]:
            prescaled_min, prescaled_max = -1.0, 1.0
        elif output_activation in ["Sigmoid", "Softmax", "GumbelSoftmax"]:
            prescaled_min, prescaled_max = 0.0, 1.0
        else:
            # For unbounded network outputs, we just return the action
            return action

        # If the action space is unbounded, we just return the action
        if low.isinf().any() or high.isinf().any():
            rescaled_action = action
        else:
            rescaled_action = low + (high - low) * (action - prescaled_min) / (
                prescaled_max - prescaled_min
            )

        return rescaled_action

    def build_network_head(self, net_config: Optional[ConfigType] = None) -> None:
        """Builds the head of the network.

        :param net_config: Configuration of the head.
        :type net_config: Optional[ConfigType]
        """
        self.head_net = self.create_mlp(
            num_inputs=self.latent_dim,
            num_outputs=spaces.flatdim(self.action_space),
            name="actor",
            net_config=net_config,
        )

    def forward(self, obs: TorchObsType) -> torch.Tensor:
        """Forward pass of the network.

        :param obs: Observation input.
        :type obs: TorchObsType
        :return: Output of the network.
        :rtype: torch.Tensor
        """
        latent = self.extract_features(obs)
        action = self.head_net(latent)

        # Action scaling only relevant for continuous action spaces
        if isinstance(self.action_space, spaces.Box) and self.clip_actions:
            action = DeterministicActor.rescale_action(
                action=action,
                low=self.action_low,
                high=self.action_high,
                output_activation=self.output_activation,
            )

        return action

    def recreate_network(self) -> None:
        """Recreates the network."""
        self.recreate_encoder()

        head_net = self.create_mlp(
            num_inputs=self.latent_dim,
            num_outputs=spaces.flatdim(self.action_space),
            name="actor",
            net_config=self.head_net.net_config,
        )

        self.head_net = EvolvableModule.preserve_parameters(self.head_net, head_net)


class StochasticActor(EvolvableNetwork):
    """Stochastic actor network for policy-gradient algorithms. Given an observation, constructs
    a distribution over the action space from the logits output by the network. Contains methods
    to sample actions and compute log probabilities and the entropy of the action distribution,
    relevant for many policy-gradient algorithms such as PPO, A2C, TRPO.

    :param observation_space: Observation space of the environment.
    :type observation_space: spaces.Space
    :param action_space: Action space of the environment
    :type action_space: spaces.Space
    :param encoder_cls: Encoder class to use for the network. Defaults to None, whereby it is
        automatically built using an AgileRL module according the observation space.
    :type encoder_cls: Optional[Union[str, Type[EvolvableModule]]]
    :param encoder_config: Configuration of the encoder network.
    :type encoder_config: ConfigType
    :param head_config: Configuration of the network MLP head.
    :type head_config: Optional[ConfigType]
    :param action_std_init: Initial log standard deviation of the action distribution. Defaults to 0.0.
    :type action_std_init: float
    :param squash_output: Whether to squash the output to the action space.
    :type squash_output: bool
    :param n_agents: Number of agents in the environment. Defaults to None, which corresponds to
        single-agent environments.
    :type n_agents: Optional[int]
    :param latent_dim: Dimension of the latent space representation.
    :type latent_dim: int
    :param simba: Whether to use the SimBa architecture for training the network.
    :type simba: bool
    :param recurrent: Whether to use a recurrent network.
    :type recurrent: bool
    :param device: Device to use for the network.
    :type device: str
    """

    head_net: EvolvableDistribution
    supported_spaces = (
        spaces.Box,
        spaces.Discrete,
        spaces.MultiDiscrete,
        spaces.MultiBinary,
    )

    def __init__(
        self,
        observation_space: spaces.Space,
        action_space: spaces.Space,
        encoder_cls: Optional[Union[str, Type[EvolvableModule]]] = None,
        encoder_config: Optional[ConfigType] = None,
        head_config: Optional[ConfigType] = None,
        action_std_init: float = 0.0,
        squash_output: bool = False,
        min_latent_dim: int = 8,
        max_latent_dim: int = 128,
        n_agents: Optional[int] = None,
        latent_dim: int = 32,
        simba: bool = False,
        recurrent: bool = False,
        device: str = "cpu",
    ):
        super().__init__(
            observation_space,
            encoder_cls=encoder_cls,
            encoder_config=encoder_config,
            action_space=action_space,
            min_latent_dim=min_latent_dim,
            max_latent_dim=max_latent_dim,
            n_agents=n_agents,
            latent_dim=latent_dim,
            simba=simba,
            recurrent=recurrent,
            device=device,
        )

        # Require the head to output logits to parameterize a distribution
        if head_config is None:
            head_config = MlpNetConfig(hidden_size=[32], output_activation=None)
        else:
            head_config["output_activation"] = None

        self.action_std_init = action_std_init
        self.squash_output = squash_output
        self.action_space = action_space

        self.build_network_head(head_config)
        self.output_activation = None

        if isinstance(self.action_space, spaces.Box):
            self.action_low = torch.as_tensor(self.action_space.low, device=self.device)
            self.action_high = torch.as_tensor(
                self.action_space.high, device=self.device
            )
        else:
            self.action_low = None
            self.action_high = None

        # Wrap the network in an EvolvableDistribution
        self.head_net = EvolvableDistribution(
            action_space=action_space,
            network=self.head_net,
            action_std_init=action_std_init,
            squash_output=squash_output,
            device=device,
        )

    def build_network_head(self, net_config: Optional[ConfigType] = None) -> None:
        """Builds the head of the network.

        :param net_config: Configuration of the head.
        :type net_config: Optional[ConfigType]
        """
        self.head_net = self.create_mlp(
            num_inputs=self.latent_dim,
            num_outputs=spaces.flatdim(self.action_space),
            name="actor",
            net_config=net_config,
        )

    def scale_action(self, action: torch.Tensor) -> torch.Tensor:
        """Scale the action to the action space.

        :param action: Action.
        :type action: torch.Tensor
        :return: Scaled action.
        :rtype: torch.Tensor
        """
        return self.action_low + (
            0.5 * (action + 1.0) * (self.action_high - self.action_low)
        )

    def forward(
        self, obs: TorchObsType, action_mask: Optional[ArrayOrTensor] = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Forward pass of the network.

        :param obs: Observation input.
        :type obs: TorchObsType
        :param action_mask: Action mask.
        :type action_mask: Optional[ArrayOrTensor]
        :return: Action and log probability of the action.
        :rtype: Tuple[torch.Tensor, torch.Tensor]
        """
        latent = self.extract_features(obs)
        action, log_prob, entropy = self.head_net.forward(latent, action_mask)

        # Action scaling only relevant for continuous action spaces with squashing
        if isinstance(self.action_space, spaces.Box) and self.squash_output:
            action = self.scale_action(action)

        return action, log_prob, entropy

    def action_log_prob(self, action: torch.Tensor) -> torch.Tensor:
        """Get the log probability of the action.

        :param action: Action.
        :type action: torch.Tensor
        :return: Log probability of the action.
        :rtype: torch.Tensor
        """
        return self.head_net.log_prob(action)

    def action_entropy(self) -> torch.Tensor:
        """Get the entropy of the action distribution.

        :return: Entropy of the action distribution.
        :rtype: torch.Tensor
        """
        return self.head_net.entropy()

    def recreate_network(self) -> None:
        """Recreates the network with the same parameters as the current network.

        :param shrink_params: Whether to shrink the parameters of the network. Defaults to False.
        :type shrink_params: bool
        """
        self.recreate_encoder()

        head_net = self.create_mlp(
            num_inputs=self.latent_dim,
            num_outputs=spaces.flatdim(self.action_space),
            name="actor",
            net_config=self.head_net.net_config,
        )

        head_net = EvolvableDistribution(
            self.action_space,
            head_net,
            action_std_init=self.action_std_init,
            squash_output=self.squash_output,
            device=self.device,
        )

        self.head_net = EvolvableModule.preserve_parameters(self.head_net, head_net)



from dataclasses import dataclass, field
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import torch
import yaml

from agilerl.typing import ConfigType


@dataclass
class NetConfig:
    """Dataclass for storing evolvable network configurations."""

    @classmethod
    def from_dict(cls, config: Dict[str, Any]) -> "NetConfig":
        return cls(**config)

    @classmethod
    def from_yaml(cls, path: str) -> "NetConfig":
        with open(path) as file:
            config: Dict[str, Any] = yaml.safe_load(file)
            assert "NET_CONFIG" in config, "NET_CONFIG not found in yaml file."
            net_config = config["NET_CONFIG"]

        return cls.from_dict(net_config)

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)

    def __setitem__(self, key: str, value: Any) -> None:
        setattr(self, key, value)

    def get(self, key: str, default: Any = None) -> Any:
        return getattr(self, key, default)

    def pop(self, key: str, default: Any = None) -> Any:
        attr = getattr(self, key, default)
        if attr is not default:
            delattr(self, key)
        return attr

    def keys(self) -> List[str]:
        return list(self.__dataclass_fields__.keys())

    def values(self) -> List[Any]:
        return [getattr(self, key) for key in self.keys()]

    def items(self) -> Dict[str, Any]:
        return {key: getattr(self, key) for key in self.keys()}.items()


@dataclass
class MlpNetConfig(NetConfig):
    hidden_size: List[int]
    activation: str = field(default="ReLU")
    output_activation: Optional[str] = field(default=None)
    min_hidden_layers: int = field(default=1)
    max_hidden_layers: int = field(default=3)
    min_mlp_nodes: int = field(default=16)
    max_mlp_nodes: int = field(default=500)
    layer_norm: bool = field(default=True)
    output_vanish: bool = field(default=True)
    init_layers: bool = field(default=True)
    noisy: bool = field(default=False)
    noise_std: float = field(default=0.5)

    def __post_init__(self):
        assert (
            len(self.hidden_size) >= self.min_hidden_layers
        ), "Hidden layers must be greater than min_hidden_layers."

        assert (
            len(self.hidden_size) <= self.max_hidden_layers
        ), "Hidden layers must be less than max_hidden_layers."

        assert all(
            [
                self.min_mlp_nodes <= nodes and nodes <= self.max_mlp_nodes
                for nodes in self.hidden_size
            ]
        ), "Nodes must be within min_nodes and max_nodes."


@dataclass
class SimBaNetConfig(NetConfig):
    hidden_size: int
    num_blocks: int
    output_activation: Optional[str] = field(default=None)
    min_blocks: int = field(default=1)
    max_blocks: int = field(default=4)
    min_mlp_nodes: int = field(default=16)
    max_mlp_nodes: int = field(default=500)

    def __post_init__(self):
        assert (
            self.num_blocks >= self.min_blocks
        ), "Number of residual blocks must be greater than min_blocks."

        assert (
            self.num_blocks <= self.max_blocks
        ), "Number of residual blocks must be less than max_blocks."

        assert (
            self.min_mlp_nodes <= self.hidden_size
            and self.hidden_size <= self.max_mlp_nodes
        ), "Nodes must be within min_nodes and max_nodes."


@dataclass
class CnnNetConfig(NetConfig):
    channel_size: List[int]
    kernel_size: List[Union[int, Tuple[int, ...]]]
    stride_size: List[int]
    sample_input: Optional[torch.Tensor] = field(default=None)
    activation: str = field(default="ReLU")
    output_activation: Optional[str] = field(default=None)
    block_type: Literal["Conv2d", "Conv3d"] = field(default="Conv2d")
    min_hidden_layers: int = field(default=1)
    max_hidden_layers: int = field(default=6)
    min_channel_size: int = field(default=32)
    max_channel_size: int = field(default=256)
    layer_norm: bool = field(default=True)
    init_layers: bool = field(default=True)


@dataclass
class LstmNetConfig(NetConfig):
    hidden_size: int
    num_layers: int = field(default=1)
    min_hidden_size: int = field(default=16)
    max_hidden_size: int = field(default=500)
    min_layers: int = field(default=1)
    max_layers: int = field(default=4)
    output_activation: Optional[str] = field(default=None)
    dropout: float = field(default=0.0)


@dataclass
class MultiInputNetConfig(NetConfig):
    """Configuration for the EvolvableMultiInput network.

    This configuration class combines settings for the multi-input network,
    including latent dimension settings and network-specific configurations.
    """

    latent_dim: int = 16
    min_latent_dim: int = 8
    max_latent_dim: int = 128
    vector_space_mlp: bool = False
    output_activation: Optional[str] = field(default=None)

    # Network configurations
    cnn_config: Optional[ConfigType] = field(default=None)
    mlp_config: Optional[ConfigType] = field(default=None)
    lstm_config: Optional[ConfigType] = field(default=None)

    # Additional settings
    init_dicts: Dict[str, Dict[str, Any]] = field(default_factory=dict)

    def __post_init__(self):
        """Validate configuration parameters after initialization."""
        # Validate latent dimension
        assert (
            self.latent_dim >= self.min_latent_dim
        ), f"Latent dimension {self.latent_dim} must be >= {self.min_latent_dim}"
        assert (
            self.latent_dim <= self.max_latent_dim
        ), f"Latent dimension {self.latent_dim} must be <= {self.max_latent_dim}"

        # Validate network configurations if provided
        if self.cnn_config is not None:
            assert isinstance(
                self.cnn_config, CnnNetConfig
            ), "CNN config must be an instance of CnnNetConfig"
        else:
            self.cnn_config = CnnNetConfig(
                channel_size=[16, 16],
                kernel_size=[3, 3],
                stride_size=[1, 1],
                output_activation="ReLU",
            )
        if self.mlp_config is not None:
            assert isinstance(
                self.mlp_config, MlpNetConfig
            ), "MLP config must be an instance of MlpNetConfig"
        else:
            self.mlp_config = MlpNetConfig(
                hidden_size=[64, 64],
                output_activation="ReLU",
            )

        if self.lstm_config is not None:
            assert isinstance(
                self.lstm_config, LstmNetConfig
            ), "LSTM config must be an instance of LstmNetConfig"
        else:
            self.lstm_config = LstmNetConfig(
                hidden_size=64,
                num_layers=2,
                output_activation="ReLU",
            )
